# coding: utf-8

"""
Module to parse BED12 objects. Much more basic than what PyBedtools could offer,
but at the same time more pythonic.
"""

from time import sleep
import os
from Bio import Seq
import Bio.SeqRecord
from .parser import Parser, _attribute_definition
from sys import intern
import copy
from ..exceptions import InvalidParsingFormat
from ..parsers.GFF import GffLine
from typing import Union
import re
import pysam
import functools
from Bio import BiopythonWarning
import warnings
from Bio.Data.IUPACData import ambiguous_dna_letters as _ambiguous_dna_letters
from Bio.Data.IUPACData import ambiguous_rna_letters as _ambiguous_rna_letters
from Bio.Data import CodonTable
import multiprocessing as mp
import msgpack
import logging
import logging.handlers as logging_handlers
from ..utilities import to_bool
from ..utilities.log_utils import create_null_logger
import pyfaidx
import zlib
import numpy as np
import random
import pprint as pp
from math import modf
from Bio.Data import IUPACData


codons = copy.deepcopy(CodonTable.ambiguous_dna_by_id[1]._codon_table)
codons.start_codons = ["ATG"]
standard = CodonTable.AmbiguousCodonTable(codons,
                                          IUPACData.ambiguous_dna_letters,
                                          IUPACData.ambiguous_dna_values,
                                          IUPACData.extended_protein_letters,
                                          IUPACData.extended_protein_values)
assert standard.start_codons == ["ATG"]
assert CodonTable.ambiguous_dna_by_id[1].start_codons != ["ATG"]

ambiguous_dna_by_id = dict()
ambiguous_dna_by_name = dict()
for key, table in CodonTable.ambiguous_dna_by_name.items():
    ambiguous_dna_by_name[key] = table

for key, table in CodonTable.ambiguous_dna_by_id.items():
    ambiguous_dna_by_id[key] = table
ambiguous_dna_by_id[0] = standard


@functools.lru_cache(typed=True, maxsize=2**10)
def get_tables(table, to_stop=False, gap=None, stop_symbol="*"):
    forward_table = table.forward_table.forward_table.copy()
    stop_codons = set(table.stop_codons)
    dual_coding = [c for c in stop_codons if c in forward_table]
    if dual_coding:
        c = dual_coding[0]
        if to_stop:
            raise ValueError("You cannot use 'to_stop=True' with this table "
                             "as it contains {} codon(s) which can be both "
                             " STOP and an  amino acid (e.g. '{}' -> '{}' or "
                             "STOP)."
                             .format(len(dual_coding), c, forward_table[c]))
        warnings.warn("This table contains {} codon(s) which code(s) for both "
                      "STOP and an amino acid (e.g. '{}' -> '{}' or STOP). "
                      "Such codons will be translated as amino acid."
                      .format(len(dual_coding), c, forward_table[c]),
                      BiopythonWarning)

    for stop in stop_codons:
        forward_table[stop] = stop_symbol
    if gap is not None:
        forward_table[gap * 3] = stop_symbol

    valid_letters = set(table.nucleotide_alphabet.upper())

    getter = np.vectorize(forward_table.get, otypes=["<U"])
    return forward_table, getter, valid_letters


def _translate_str(sequence, table, stop_symbol="*", to_stop=False, cds=False, pos_stop="X", gap=None):
    """Translate nucleotide string into a protein string (PRIVATE).

    Arguments:
     - sequence - a string
     - table - a CodonTable object (NOT a table name or id number)
     - stop_symbol - a single character string, what to use for terminators.
     - to_stop - boolean, should translation terminate at the first
       in frame stop codon?  If there is no in-frame stop codon
       then translation continues to the end.
     - pos_stop - a single character string for a possible stop codon
       (e.g. TAN or NNN)
     - cds - Boolean, indicates this is a complete CDS.  If True, this
       checks the sequence starts with a valid alternative start
       codon (which will be translated as methionine, M), that the
       sequence length is a multiple of three, and that there is a
       single in frame stop codon at the end (this will be excluded
       from the protein sequence, regardless of the to_stop option).
       If these tests fail, an exception is raised.
     - gap - Single character string to denote symbol used for gaps.
       Defaults to None.

    Returns a string.

    e.g.

    >>> from Bio.Data import CodonTable
    >>> table = CodonTable.ambiguous_dna_by_id[1]
    >>> _translate_str("AAA", table)
    'K'
    >>> _translate_str("TAR", table)
    '*'
    >>> _translate_str("TAN", table)
    'X'
    >>> _translate_str("TAN", table, pos_stop="@")
    '@'
    >>> _translate_str("TA?", table)
    Traceback (most recent call last):
       ...
    Bio.Data.CodonTable.TranslationError: Codon 'TA?' is invalid

    In a change to older versions of Biopython, partial codons are now
    always regarded as an error (previously only checked if cds=True)
    and will trigger a warning (likely to become an exception in a
    future release).

    If **cds=True**, the start and stop codons are checked, and the start
    codon will be translated at methionine. The sequence must be an
    while number of codons.

    >>> _translate_str("ATGCCCTAG", table, cds=True)
    'MP'
    >>> _translate_str("AAACCCTAG", table, cds=True)
    Traceback (most recent call last):
       ...
    Bio.Data.CodonTable.TranslationError: First codon 'AAA' is not a start codon
    >>> _translate_str("ATGCCCTAGCCCTAG", table, cds=True)
    Traceback (most recent call last):
       ...
    Bio.Data.CodonTable.TranslationError: Extra in frame stop codon found.
    """

    # Check that the pos_stop is a single character
    # By default this is the "X" character (equivalent to "N" for nucleotides)
    if not (isinstance(pos_stop, (bytes, str)) and len(pos_stop) == 1):
        raise ValueError("Pos_stop must be a single character, not {pos_stop}".format(pos_stop=pos_stop))
    if isinstance(pos_stop, bytes):
        pos_stop = pos_stop.decode()

    if cds and len(sequence) % 3 != 0:
        raise CodonTable.TranslationError("Sequence length {0} is not a multiple of three".format(
            len(sequence)
        ))
    elif gap is not None and (not isinstance(gap, str) or len(gap) > 1):
        raise ValueError("Gap character should be a single character string.")

    forward_table, getter, valid_letters = get_tables(table, to_stop=to_stop, gap=gap, stop_symbol=stop_symbol)

    sequence = sequence.upper()
    if not valid_letters.issuperset(set(sequence)):
        raise CodonTable.TranslationError("Invalid letters in the sequence: {}".format(
            set.difference(set(sequence), valid_letters)
        ))

    amino_acids = getter(np.array(
        [sequence[start:start + 3] for start in range(0, len(sequence) - len(sequence) % 3, 3)], dtype="<U"))

    if cds and amino_acids[0] != "M":
        if sequence[0:3] in table.start_codons:
            amino_acids[0] = "M"
        else:
            raise CodonTable.TranslationError("First codon '{0}' is not a start codon".format(sequence[:3]))

    nones = np.where((amino_acids == "None") | (amino_acids is None))[0]
    amino_acids[nones] = pos_stop
    _stop_locations = np.where(amino_acids == stop_symbol)[0]
    found_stops = _stop_locations.shape[0]

    if cds and found_stops > 1:
        raise CodonTable.TranslationError(
            "Extra in-frame stop codon found. Sequence: {sequence}".format(sequence=sequence))
    elif cds and found_stops and _stop_locations[0] < len(amino_acids) - 1:
        raise CodonTable.TranslationError(
            "Extra in-frame stop codon. Sequence:\n{sequence}\n{spaces}^^^".format(
            sequence=sequence, spaces=" " * _stop_locations[0] * 3))
    if to_stop and found_stops > 0:
        amino_acids = amino_acids[:_stop_locations[0]]

    return "".join(amino_acids)


# These classes do contain lots of things, it is correct like it is
# pylint: disable=too-many-instance-attributes
class BED12:

    """
    BED12 parsing class.
    """

    __valid_coding = {"True": True, "False": False, True: True, False: False}

    _attribute_pattern = re.compile(r"([^;]*)=([^$=]*)(?:;|$)")

    def __init__(self, *args: Union[str, list, tuple, GffLine],
                 fasta_index=None,
                 phase=None,
                 sequence=None,
                 transcriptomic=False,
                 max_regression=0,
                 start_adjustment=True,
                 coding=True,
                 lenient=False,
                 table=0,
                 logger=create_null_logger()):

        """
        :param args: the BED12 line.
        :type args: (str, list, tuple, GffLine)

        :param fasta_index: Optional FAI index
        :param sequence: Optional sequence string

        :param transcriptomic: boolean flag
        :type transcriptomic: bool

        :param lenient: boolean flag. If set as True, the BED12 object will skip certain checks on the validity of the
                        CDS (e.g. the CDS length being required to be a multiple of three, after accounting for the
                        phase).
        :type lenient: bool

        Constructor method.

        Each instance will have:

        :param chrom: chromosome
        :type chrom: str

        :param header: whether the line is a header line or not
        :type header: bool

        :param start: start position
        :type start: int

        :param end: end position
        :type end: int

        :param name: Name of the feature
        :type name: str

        :param score: score assigned to the feature.
        :type score: float
        :type score: None

        :param strand of the feature

        :param thickStart: "internal" start of the feature (e.g. CDS start)
        :type thickStart: int

        :param thickEnd: "internal" end of the feature (e.g. CDS end)
        :type thickEnd: int

        :param rgb: RGB color scheme. Currently not checked
        :type rgb: str

        :param blockCount: number of blocks (e.g. exons)
        :type blockCount: int

        :param blockSizes: sizes of the blocks (e.g. exons). Its length must be equal to blockCount
        :type blockSizes: list(int)

        :param blockStarts: list of start positions for the blocks.
        Its length must be equal to blockCount
        :type blockStarts: list(int)

        Additional parameters calculated inside the class:

        :param fasta_length: length of the feature (see len() method)
        :type fasta_length: int

        :param has_start_codon: flag. For transcriptomic BED12, it indicates
        whether we have a start codon or not.
        :type has_start_codon: bool

        :param has_stop_codon: flag. For transcriptomic BED12, it indicates
        whether we have a stop codon or not.
        :type has_stop_codon: bool

        :param start_codon: string of the start codon, if found
        :type start_codon: None
        :type start_codon: str

        :param stop_codon: string of the stop codon, if found
        :type stop_codon: None
        :type stop_codon: str
        """

        self._line = None
        self.header = False
        self.__phase = None  # Initialize to None for the non-transcriptomic objects
        self.__has_start = False
        self.__has_stop = False
        self.__transcriptomic = False
        self.__parent = None
        self.transcriptomic = transcriptomic
        if phase is not None:
            self.phase = phase  # This will check the validity of the phase itself
        self.__strand = None
        self.__max_regression = 0
        # This value will be set when checking for the internal sequence
        # If >=1, i.e. at least one internal stop codon, the ORF is invalid
        self._internal_stop_codons = 0
        self.chrom = None
        self.__start = self.__end = self.__thick_start = self.__thick_end = 0
        self.name = ""
        self.score = 0
        self.strand = None
        self.rgb = ''
        self.stop_codon = self.start_codon = None
        self.__has_start = self.__has_stop = False
        self.__block_sizes = np.zeros(1, dtype=np.int64)
        self.__block_starts = np.zeros(1, dtype=np.int64)
        self.__block_count = 1
        self.__invalid = None
        self.invalid_reason = None
        self.fasta_length = None
        self.__in_index = True
        self.max_regression = max_regression
        self.start_adjustment = start_adjustment
        self.coding = coding
        if self.coding and self.phase is None:
            self.phase = 0
        self.__table = standard
        self.__table_index = 0
        self.table = table
        self.__lenient = lenient
        self.alias = None
        self.__logger = create_null_logger()
        self.logger = logger
        self.attributes = dict()
        self.logger.debug("Set the basic properties for %s", self.chrom)

        if len(args) == 0:
            self.header = True
            return

        self._line = args[0]
        if isinstance(self._line, str) or self._line is None:
            if self._line is None:
                self._line = ''
            self._line = self._line.rstrip()
            if len(self._line) == 0 or self._line[0] == "#":
                self.header = True
                return
            self._fields = self._line.split("\t")
            if len(self._fields) in (12, 13):
                self.__set_values_from_fields()
                self.header = False
            else:
                self.header = True
                return
        elif isinstance(self._line, type(self)):  # Re-initialising with another object
            self.__set_values_from_bed12(args[0])
        elif isinstance(self._line, GffLine):
            if self._line.header is True:
                self.header = True
                return
            elif self.transcriptomic is False:
                error = "GFF lines can be used as BED12-equivalents only in a transcriptomic context."
                self.logger.error(error)
                raise InvalidParsingFormat(error)
            else:
                self.header = False
                if sequence:
                    fasta_length = len(sequence)
                elif fasta_index:
                    if isinstance(fasta_index, pysam.FastaFile):
                        fasta_length = fasta_index.get_reference_length(self._line.chrom)
                    elif isinstance(fasta_index, pyfaidx.Fasta):
                        sequence = fasta_index[self._line.chrom]
                        fasta_length = len(fasta_index[self._line.chrom])
                    else:
                        raise TypeError("Invalid FASTA index")
                else:
                    raise ValueError("Either a sequence or a FAI index are needed")
                self.__set_values_from_gff(fasta_length)

        elif not (isinstance(self._line, list) or isinstance(self._line, tuple)):
            raise InvalidParsingFormat("I need an ordered array, not {0}".format(type(self._line)))
        else:
            self._fields = self._line
            print("Line", self._fields)
            self.__set_values_from_fields()

        self.__check_validity(transcriptomic, fasta_index, sequence)

        if self.invalid and self.coding:
            self.logger.debug("%s cannot be coding as it is invalid (reason: %s)", self.chrom, self.invalid_reason)
            self.coding = False

        if self.coding and self.phase is None:
            self.phase = 0

    @property
    def is_transcript(self):
        """BED12 files are always transcripts for Mikado."""
        return True

    @property
    def is_gene(self):
        """BED12 files are never "genes" according to Mikado"""
        return False

    @property
    def source(self):
        return "bed12"

    @property
    def gene(self):
        return self.parent[0] if self.parent else None

    @property
    def parent(self):
        return self.__parent

    @property
    def table(self):
        return self.__table

    @table.setter
    def table(self, table):
        if isinstance(table, bool):  # Boolean can be considered as int so this requires special handling
            raise ValueError(f"Invalid table specified: {table} (type {type(table)})")
        elif table is not None and not isinstance(table, (int, float, bytes, str)):
            raise ValueError(f"Invalid table specified: {table} (type {type(table)})")
        elif isinstance(table, (str, bytes)):
            table = table.decode() if isinstance(table, bytes) else table
            if table.isdigit() is True:
                table = int(table)
            elif re.search(r"^[0-9]*\.[0-9]$", table):
                table = float(table)
                if modf(table) != 0:
                    raise ValueError(f"Invalid table specified: {table}")
                table = int(table)
        elif isinstance(table, float):
            if modf(table) != 0:
                raise ValueError(f"Invalid table specified: {table}")
            table = int(table)

        if table is None:
            self.__table = standard
            self.__table_index = 0
        elif isinstance(table, int):
            if table not in ambiguous_dna_by_id.keys():
                raise ValueError(f"Invalid table code specified: {table}. Available codes: "
                                 f"{', '.join([str(_) for _ in ambiguous_dna_by_id.keys()])}")
            self.__table = ambiguous_dna_by_id[table]
            assert self.__table.start_codons == ["ATG"] if table == 0 else True, f"Invalid codons for table 0: " \
                                                                                 f"{self.__table.start_codons}"
            self.__table_index = table
        elif isinstance(table, str):
            if table not in ambiguous_dna_by_name.keys():
                raise ValueError(f"Invalid table name specified: {table}. Available table: "
                                 f"{', '.join([str(_) for _ in ambiguous_dna_by_name.keys()])}")
            self.__table = ambiguous_dna_by_name[table]
            self.__table_index = ambiguous_dna_by_name[table].id
        return

    @parent.setter
    def parent(self, parent):
        if parent is not None and not isinstance(parent, str):
            raise TypeError(type(parent))
        self.__parent = [parent]

    def __getstate__(self):

        state = copy.deepcopy(dict((key, val) for key, val in self.__dict__.items()
                                   if key not in ("_BED12_table") and
                                   not isinstance(val, logging.Logger) and
                                   not isinstance(val, CodonTable.CodonTable)))

        return state

    def __setstate__(self, state):
        # del state["table"]
        self.__dict__.update(state)
        self.table = self.__table_index

    def _parse_attributes(self, attributes):

        """
        Private method that parses the last field of the GFF line.
        :return:
        """

        self.attribute_order = []

        infolist = self._attribute_pattern.findall(attributes.rstrip().rstrip(";"))

        for item in infolist:
            key, val = item
            if key.lower() in ("parent", "geneid"):
                self.parent = val
            elif "phase" in key.lower():
                self.phase = int(val)
                self.coding = True
            elif key.lower() == "coding":
                self.coding = self.__valid_coding.get(val, False)
                if self.transcriptomic is True:
                    self.phase = 0
            elif key.lower() == "alias":
                self.alias = val
            elif key.lower() == "id":
                self.name = val
            else:
                self.attributes[key.capitalize()] = _attribute_definition(val)
                continue

    def __set_values_from_fields(self):

        """
        Private method that sets the correct values from the fields derived from the input line.
        :return:
        """
        self.chrom, self.start, self.end, \
            self.name, self.score, self.strand, \
            self.thick_start, self.thick_end, self.rgb, \
            self.block_count, block_sizes, block_starts = self._fields[:12]

        # Reduce memory usage
        intern(self.chrom)
        self.start = int(self.start) + 1
        self.end = int(self.end)
        try:
            self.score = float(self.score)
        except (ValueError, TypeError):
            self.score = None
        self.thick_start = int(self.thick_start) + 1
        self.thick_end = int(self.thick_end)
        self.block_count = int(self.block_count)
        if isinstance(block_sizes, (str, bytes)):
            self.block_sizes = [int(x) for x in block_sizes.split(",") if x]
        else:
            self.block_sizes = [int(x) for x in block_sizes]
        if isinstance(block_starts, (str, bytes)):
            self.block_starts = [int(x) for x in block_starts.split(",") if x]
        else:
            self.block_starts = [int(x) for x in block_starts]
        to_parse = self.name
        if len(self._fields) == 13:
            to_parse = ";".join([to_parse, self._fields[-1]])
        self._parse_attributes(to_parse)
        self.has_start_codon = False
        self.has_stop_codon = False
        self.start_codon = None
        self.stop_codon = None
        self.fasta_length = len(self)
        return

    def __set_values_from_bed12(self, line):

        self.__setstate__(line.__getstate__())
        return

    def __set_values_from_gff(self, fasta_length):
        """
        Private method that sets the correct values from the fields derived from an input GFF line.
        :return:
        """

        (self.chrom, self.thick_start,
         self.thick_end, self.strand, self.name) = (self._line.chrom,
                                                    self._line.start,
                                                    self._line.end, self._line.strand, self._line.id)
        intern(self.chrom)
        if self.name is None:
            raise InvalidParsingFormat("{self} should have the name property defined".format(self=repr(self)))
        self.start = 1
        self.end = fasta_length
        self.score = self._line.score
        self.rgb = None
        self.block_count = 1
        self.block_sizes = [self.thick_end - self.thick_start +1]
        self.block_starts = [self.thick_start]
        self.has_start_codon = False
        self.has_stop_codon = False
        self.start_codon = None
        self.stop_codon = None
        self.fasta_length = fasta_length
        return

    def __check_validity(self, transcriptomic, fasta_index, sequence):
        """
        Private method that checks that the BED12 object has been instantiated correctly.

        :return:
        """

        del self.invalid

        if transcriptomic is True and self.coding is True:
            if not (fasta_index is not None or sequence is not None):
                self.logger.debug("No further check on the validity of %s as no sequence has been provided.",
                                  self.chrom)
                return

        if transcriptomic is True:
            self.has_start_codon = False
            self.has_stop_codon = False

        if transcriptomic is True and self.coding is True and (fasta_index is not None or sequence is not None):
            self.logger.debug("Starting to check the validity of %s", self.chrom)
            self.validity_checked = True
            if sequence is not None:
                self.fasta_length = len(sequence)
                if hasattr(sequence, "seq"):
                    sequence = str(sequence.seq)
                if not isinstance(sequence, str):
                    sequence = str(sequence)
            else:
                if self.id not in fasta_index:
                    self.logger.warning("%s not found in the index. Aborting the check, we will trust the ORF as-is.")
                    self.__in_index = False
                    return
                self.fasta_length = len(fasta_index[self.id])
                sequence = fasta_index[self.id]
                if hasattr(sequence, "seq"):
                    sequence = str(sequence.seq)
                if not isinstance(sequence, str):
                    sequence = str(sequence)

            assert isinstance(sequence, str)
            # Just double check that the sequence length is the same as what the BED would suggest
            if self.__is_invalid() is True:
                self.logger.debug("%s is invalid (%s)", self.chrom, self.invalid_reason)
                self.coding = False
                return

            if self.strand != "-":
                orf_sequence = sequence[
                               (self.thick_start - 1 if not self.phase else self.start + self.phase - 1):self.thick_end]
            else:
                orf_sequence = Seq.reverse_complement(
                    sequence[(self.thick_start - 1):(
                        self.thick_end if not self.phase else self.end - (3 - self.phase) % 3)])

            self.start_codon = str(orf_sequence)[:3].upper()
            self.stop_codon = str(orf_sequence[-3:]).upper()

            if self.start_codon in self.table.start_codons and (self.phase is None or self.phase == 0):
                self.logger.debug("Found start codon for %s. Setting phase to 0", self.chrom)
                self.has_start_codon = True
                self.phase = 0
            else:
                self.has_start_codon = False
                if self.start_adjustment is True:
                    self._adjust_start(sequence, orf_sequence)

            if self.stop_codon in self.table.stop_codons:
                self.has_stop_codon = True
            else:
                self.has_stop_codon = False
                if self.strand == "+" and self.end - self.thick_end < 3:
                    self.thick_end = self.end
                elif self.strand == "-" and self.thick_start - self.start < 3:
                    self.thick_start = 1

            self.logger.debug("%s with start codon (%s) and stop codon (%s). Valid: %s",
                              self.chrom, self.has_start_codon, self.has_stop_codon, not self.invalid)

            # Get only a proper multiple of three
            if self.lenient is False and self.coding is True:
                if self.strand != "-":
                    orf_sequence = sequence[
                                   (self.thick_start - 1 if not self.phase
                                    else self.start + self.phase - 1):self.thick_end]
                else:
                    orf_sequence = Seq.reverse_complement(
                        sequence[
                        (self.thick_start - 1):
                        (self.thick_end if not self.phase else self.end - self.phase)])

                last_pos = -3 - ((len(orf_sequence)) % 3)
                translated_seq = _translate_str(orf_sequence[:last_pos],
                                                table=self.table,
                                                gap='N')

                self._internal_stop_codons = str(translated_seq).count("*")
                if self._internal_stop_codons == 0 and len(orf_sequence[last_pos:]) > 3:
                    if orf_sequence[last_pos:][:3] in self.table.stop_codons:
                        # We need to shift either the thick start or the thick end. Note that last_pos is negative
                        # so we need to minus it.
                        if self.strand == "-":
                            self.thick_start += -last_pos % 3
                            self.logger.warning(
                                f"Shifting the position of the thick start of {self.name} by {-last_pos % 3}")
                        else:
                            self.thick_end -= -last_pos % 3
                            self.logger.warning(
                                f"Shifting the position of the thick end of {self.name} by {-last_pos % 3}")

            del self.invalid
            if self.__is_invalid() is True:
                return

    def _adjust_start(self, sequence, orf_sequence):

        if len(orf_sequence) != (self.thick_end - self.thick_start + 1 - self.phase):
            # We are checking that the sequence of the ORF (provided as argument) is the same length as
            # the imputed length of the ORF
            raise ValueError("The provided orf_sequence of length {lorf} is different from the imputed length of the\
 ORF for {sid} (total {ltotal}; thick start {self.thick_start}, thick end {self.thick_end}, phase {self.phase})".format(
                lorf=len(orf_sequence), ltotal=self.thick_end - self.thick_start + 1 - self.phase,
                sid=self.name, self=self))

        # Let's check UPstream first.
        # This means that we DO NOT have a starting Met and yet we are starting far upstream.
        # self.logger.debug("Starting to adjust the start of %s", self.chrom)
        if self.strand == "+" and self.thick_start > 3:
            for pos in range(self.thick_start, 3, -3):
                self.thick_start -= 3
                codon = sequence[pos - 4:pos - 1]
                is_start, is_stop = ((codon in self.table.start_codons),
                                     (codon in self.table.stop_codons))
                self.logger.debug("Checking pos %s (%s) for %s, start: %s; stop: %s",
                                  pos, codon, self.chrom, is_start, is_stop)
                if is_start:
                    # We have found a valid methionine.
                    break
                elif is_stop:
                    self.stop_codon = codon
                    self._internal_stop_codons = 1
                    assert self.invalid is True
                    self.logger.debug(
                        "Found in-frame stop codon for %s while expanding, stopping here. Invalid: %s (reason %s)",
                                        self.chrom, self.invalid, self.invalid_reason)
                    break
                continue

        elif self.strand == "-" and self.end - self.thick_end > 3:
            for pos in range(self.thick_end, self.end - 3, 3):
                self.thick_end += 3
                codon = Seq.reverse_complement(sequence[pos - 3:pos])
                is_start, is_stop = ((codon in self.table.start_codons),
                                     (codon in self.table.stop_codons))
                if is_start:
                    # We have found a valid methionine.
                    self.logger.debug("Found correct start codon for %s while expanding, stopping here.",
                                      self.chrom)
                    break
                elif is_stop:
                    self.stop_codon = codon
                    self._internal_stop_codons = 1
                    assert self.invalid is True
                    self.logger.debug(
                        "Found in-frame stop codon for %s while expanding, stopping here. Invalid: %s (reason %s)",
                        self.chrom, self.invalid, self.invalid_reason)
                    break
        else:
            self._regression(orf_sequence)

        if self.has_start_codon is False:
            # The validity will be automatically checked
            # self.logger.debug("Making adjustments in %s for missing start codon", self.chrom)
            if self.strand == "+":
                if self.thick_start - self.start <= 2:
                    new_phase = max(self.thick_start - self.start, 0)
                    self.phase = new_phase
                    self.thick_start = self.start
                else:
                    self.phase = 0
            else:
                if self.end - self.thick_end <= 2:
                    new_phase = max(self.end - self.thick_end, 0)
                    self.phase = new_phase
                    self.thick_end = self.end
                else:
                    self.phase = 0
        else:
            self.logger.debug("Setting phase of %s at 0 (end: %s; thick end: %s; thick start %s)",
                              self.chrom, self.end, self.thick_end, self.thick_start)
            self.phase = 0

        del self.invalid
        if self.invalid:
            self.logger.debug("%s is not coding after checking. Reason: %s", self.chrom, self.invalid_reason)
            self.coding = False

    def _regression(self, orf_sequence):
        self.logger.debug(
            "Starting the regression algorithm to find an internal start for %s (end: %s; thick start/end: %s, %s; phase %s)",
            self.chrom, self.end, self.thick_start, self.thick_end, self.phase)
        if self.strand != "-":
            # self.thick_start = self.phase + 3
            self.logger.debug("Starting to analyse %s; positions %s-%s",
                              self.chrom,
                              self.phase + 3,
                              self.phase + 3 + int(len(orf_sequence) * self.max_regression),
                              )
            for pos in range(self.phase + 3,
                             int(len(orf_sequence) * self.max_regression),
                             3):
                codon = orf_sequence[pos:pos + 3]
                # self.logger.debug("Testing position %s-%s (%s)", pos, pos + 3, codon)
                if codon in self.table.start_codons:
                    # Now we have to shift the start accordingly
                    self.has_start_codon = True
                    self.thick_start += pos
                    self.phase = 0
                    break
                else:
                    continue
            self.logger.debug("Final internal coords for %s: %s-%s", self.chrom, self.thick_start, self.thick_end)
        elif self.strand == "-":
            if self.end - self.thick_end < 3:
                self.phase = (3 - (self.end - self.thick_end) % 3) % 3
            self.logger.debug("Starting to analyse %s (phase %s); positions %s-%s",
                              self.chrom,
                              self.phase,
                              self.phase + 3,
                              self.phase + 3 + int(len(orf_sequence) * self.max_regression),
                              )
            for pos in range(self.phase + 3,
                             int(len(orf_sequence) * self.max_regression),
                             3):
                codon = orf_sequence[pos:pos + 3]
                # self.logger.debug("Testing position %s-%s (%s)", pos, pos + 3, codon)
                if codon in self.table.start_codons:
                    # Now we have to shift the start accordingly
                    self.has_start_codon = True
                    self.thick_end -= pos
                    self.phase = 0
                    break
            self.logger.debug("Final internal coords for %s: %s-%s", self.chrom, self.thick_start, self.thick_end)

    def __repr__(self):
        return pp.saferepr(self.__dict__)

    def __hash__(self):
        return hash(frozenset(self.__getstate__()))

    def __str__(self):

        if self.header is True:
            if self._line is not None:
                return self._line
            else:
                return "#"

        line = [self.chrom, self.start - 1, self.end]
        name = "ID={};coding={}".format(self.id, self.coding)
        if self.coding:
            name += ";phase={}".format(self.phase)
        if self.alias is not None and self.alias != self.id:
            name += ";alias={}".format(self.alias)
        line.append(name)
        if not self.score:
            line.append(0)
        else:
            line.append(self.score)
        if self.strand is None:
            line.append(".")
        else:
            line.append(self.strand)
        line.extend([self.thick_start - 1, self.thick_end])
        if not self.rgb:
            line.append(0)
        else:
            line.append(self.rgb)
        line.append(self.block_count)
        line.append(",".join([str(x) for x in self.block_sizes]))
        line.append(",".join([str(x) for x in self.block_starts]))
        attributes = dict((key.lower(), val) for key, val in self.attributes.items() if key.lower() not in
                          ("geneid", "gene_id", "name", "phase", "coding", "alias", "id"))
        if self.parent is not None:
            attributes["Parent"] = self.parent[0]
            assert "Parent" in attributes
        elif "parent" in attributes:
            parent = attributes.pop("parent")
            if isinstance(parent, (tuple, list)) and len(parent) > 0:
                parent = ",".join(parent)
            elif isinstance(parent, (tuple, list)) and len(parent) == 0:
                parent = None
            attributes["Parent"] = parent
        if attributes:
            line.append(";".join(f"{key}={val}" for key, val in attributes.items() if val is not None))
        return "\t".join([str(x) for x in line])

    def __eq__(self, other):
        for key in ["chrom", "strand", "start",
                    "end", "thick_start", "thick_end",
                    "block_count", "block_sizes",
                    "block_starts"]:
            if getattr(self, key) != getattr(other, key):
                return False
        return True

    def __len__(self):
        return self.end - self.start + 1

    def copy(self):

        return copy.deepcopy(self)

    def as_simple_dict(self):

        return {
            "chrom": self.chrom,
            "id": self.id,
            "start": self.start,
            "end": self.end,
            "orf_name": self.name,
            "strand": self.strand,
            "thick_start": self.thick_start,
            "thick_end": self.thick_end,
            "score": self.score,
            "has_start_codon": self.has_start_codon,
            "has_stop_codon": self.has_stop_codon,
            "cds_len": self.cds_len,
            "phase": self.phase,
            "transcriptomic": self.transcriptomic,
        }

    @property
    def strand(self):
        """
        Strand of the feature. It must be one of None,+,-
        :rtype None | str
        """
        return self.__strand

    @strand.setter
    def strand(self, strand: str):
        """
        Setter for strand. It verifies that the value is correct.
        :param strand: New strand value
        :type strand: str | None
        """

        if strand in (".", "?", None):
            self.__strand = None
        elif strand in ("+", "-"):
            self.__strand = strand
        else:
            raise ValueError("Erroneous strand provided: {0}".format(self.strand))

    @property
    def cds_len(self):
        """
        Return the length of the internal feature i.e. the CDS:
        thickEnd-thickStart+1

        :rtype int
        """
        return self.thick_end - self.thick_start + 1

    @property
    def has_start_codon(self):
        """
        Property. True if the interval contains a start codon.
        :rtype bool
        :rtype None
        """

        return self.__has_start

    @has_start_codon.setter
    def has_start_codon(self, value: bool):
        """
        Setter for has_stop_codon. Valid values are boolean or None
        :param value: boolean flag
        :type value: bool
        :type value: None
        """

        self.__has_start = to_bool(value)

    @property
    def has_stop_codon(self):
        """
        Property. True if the interval contains a termination codon.
        :rtype bool
        :rtype None
        """
        return self.__has_stop

    @has_stop_codon.setter
    def has_stop_codon(self, value: bool):
        """
        Setter for has_stop_codon. Valid values are boolean.
        :param value: boolean flag
        :type value: bool
        :type value: None
        """
        self.__has_stop = to_bool(value)

    @property
    def full_orf(self):
        """
        Property. True if the BED12 is transcriptomic and has
        both start and stop codon, False otherwise.
        :rtype bool
        """
        return self.has_stop_codon and self.has_start_codon

    # pylint: disable=invalid-name
    @property
    def id(self):
        """
        Property. It returns the name of the feature.
        :rtype str
        """

        if self.transcriptomic is True:
            return self.chrom
        else:
            return self.name
    # pylint: enable=invalid-name

    @property
    def invalid(self):
        """
        Property. It performs basic checks on the BED line to verify its integrity.
        :rtype bool
        """

        if self.__invalid is None:
            self.__invalid = self.__is_invalid()

        return self.__invalid

    @invalid.deleter
    def invalid(self):
        self.__invalid = None

    @property
    def lenient(self):
        return self.__lenient

    def __is_invalid(self):

        if self._internal_stop_codons >= 1:
            self.invalid_reason = "{} internal stop codons found".format(self._internal_stop_codons)
            return True

        if self.fasta_length is None:
            self.fasta_length = len(self)

        if self.thick_start < self.start or self.thick_end > self.end:
            if self.thick_start == self.thick_end == self.block_sizes[0] == 0:
                pass
            else:
                invalid = "thickStart {0} <start {1}: {2}; end {3} <thickEnd {4} {5}"
                self.invalid_reason = invalid.format(self.thick_start,
                                                     self.start,
                                                     self.thick_start < self.start,
                                                     self.end,
                                                     self.thick_end,
                                                     self.thick_end > self.end)
                return True

        if self.transcriptomic is True:
            if self.__in_index is False:
                self.invalid_reason = "{} not found in the index!".format(self.chrom)
                return True

            if len(self) != self.fasta_length:
                self.invalid_reason = "FASTA length != BED length: {0} vs. {1}".format(
                    self.fasta_length,
                    len(self)
                )
                return True

            if self.lenient is True:
                pass
            else:
                if (self.cds_len - self.phase) % 3 != 0:
                    if self.strand == "+" and self.thick_end != self.end:
                        self.invalid_reason = "Invalid CDS length: {0} % 3 = {1} ({2}-{3}, {4})".format(
                            self.cds_len - self.phase,
                            (self.cds_len - self.phase) % 3,
                            self.thick_start, self.thick_end, self.phase)
                        return True
                    elif self.strand == "-" and self.thick_start != self.start:
                        self.invalid_reason = "Invalid CDS length: {0} % 3 = {1} ({2}-{3}, {4})".format(
                            self.cds_len - self.phase,
                            (self.cds_len - self.phase) % 3,
                            self.thick_start, self.thick_end, self.phase)
                        return True

        self.invalid_reason = ''
        return False

    @property
    def transcriptomic(self):
        """
        Flag. If set to True, it indicates the BED contains
        transcriptomic rather than genomic coordinates.
        :rtype bool
        """
        return self.__transcriptomic

    @transcriptomic.setter
    def transcriptomic(self, value):
        """
        Setter for transcriptomic. A valid value must be boolean.
        :type value: bool
        """

        if not isinstance(value, bool):
            raise ValueError("Invalid value: {0}".format(value))
        self.__transcriptomic = value
        if value and self.phase is None:
            self.phase = 0
        elif not value:
            self.phase = None

    @property
    def start(self):
        return self.__start

    @start.setter
    def start(self, value):
        try:
            value = int(value)
        except (ValueError, TypeError):
            raise ValueError("Start must be an integer, not {}! Value: {}".format(type(value), value))
        self.__start = value
        del self.invalid

    @start.deleter
    def start(self):
        self.__start = 0
        del self.invalid

    @property
    def end(self):
        return self.__end

    @end.setter
    def end(self, value):
        try:
            value = int(value)
        except (ValueError, TypeError):
            raise ValueError("End must be an integer, not {}! Value: {}".format(type(value), value))
        self.__end = value
        del self.invalid

    @end.deleter
    def end(self):
        self.__end = 0
        del self.invalid

    @property
    def thick_start(self):
        return self.__thick_start

    @thick_start.setter
    def thick_start(self, value):
        try:
            value = int(value)
        except (ValueError, TypeError):
            raise ValueError("Thick start must be an integer, not {}! Value: {}".format(type(value), value))
        self.__thick_start = value
        del self.invalid

    @thick_start.deleter
    def thick_start(self):
        self.__thick_start = 0
        del self.invalid

    @property
    def thick_end(self):
        return self.__thick_end

    @thick_end.setter
    def thick_end(self, value):
        try:
            value = int(value)
        except (ValueError, TypeError):
            raise ValueError("Thick end must be an integer, not {}! Value: {}".format(type(value), value))
        self.__thick_end = value
        del self.invalid

    @thick_end.deleter
    def thick_end(self):
        self.__thick_end = 0
        del self.invalid

    @property
    def phase(self):
        """This property is used for transcriptomic BED objects
        and indicates what the phase of the transcript is.
        So a BED object with an open 5'ORF whose first codon
        starts at the 1st base would have frame 1, at the second
        frame 2. In all other cases, the frame has to be 0.
        If the BED object is not transcriptomic, its frame is null
        (None).
        """

        return self.__phase

    @phase.setter
    def phase(self, val):

        if val not in (None, 0, 1, 2):
            raise ValueError("Invalid frame specified for {}: {}. Must be None or 0, 1, 2".format(
                self.name, val))
        elif self.transcriptomic is True and val not in (0, 1, 2):
            raise ValueError("A transcriptomic BED cannot have null frame.")
        del self.invalid
        self.__phase = val

    @phase.deleter
    def phase(self):
        self.__phase = None
        del self.invalid

    @property
    def block_count(self):
        return self.__block_count

    @block_count.setter
    def block_count(self, value):
        try:
            value = int(value)
        except (ValueError, TypeError):
            raise ValueError("Block count must be an integer, not {}! Value: {}".format(type(value), value))
        self.__block_count = value
        del self.invalid

    @property
    def block_sizes(self):
        return self.__block_sizes

    @block_sizes.setter
    def block_sizes(self, sizes):
        sizes = np.array(sizes)
        if not issubclass(sizes.dtype.type, np.int64):
            raise TypeError("Block sizes should be integers!")
        self.__block_sizes = sizes
        del self.invalid

    @block_sizes.deleter
    def block_sizes(self):
        self.__block_sizes = np.zeros(1, dtype=np.int64)
        del self.invalid

    @property
    def block_starts(self):
        return self.__block_starts

    @block_starts.setter
    def block_starts(self, starts):
        starts = np.array(starts)
        if not issubclass(starts.dtype.type, np.int64):
            raise TypeError("Block sizes should be integers! Dtype: {}; array: {}".format(
                starts.dtype, starts
            ))
        self.__block_starts = starts
        del self.invalid

    @block_starts.deleter
    def block_starts(self):
        self.__block_starts = np.zeros(1, dtype=np.int64)
        del self.invalid

    @property
    def _max_regression(self):
        """
        This property is used to indicate how far downstream we should go in the
          FASTA sequence to find a valid start codon, in terms of percentage
          of the the cDNA sequence. So eg in a 300 nt cDNA a max_regression value
          of 0.3 would instruct the class to look only for the first 90 bps for
          a Met.
        """

        return self.__max_regression

    @_max_regression.setter
    def _max_regression(self, value):
        if not (isinstance(value, (int, float)) and 0 <= value <= 1):
            raise ValueError(
                "Invalid value specified for _max_regression (must be between 0 and 1): {}".format(value))
        self.__max_regression = value

    def expand(self, sequence, upstream, downstream, expand_orf=False, logger=create_null_logger()):

        # TODO this needs revising. The expand_orf key does not act as it should, as the thick start and
        #  end are changed even when it is set to False.

        """This method will expand the coordinates of a BED **transcriptomic** object, and will also
        optionally try to expand the associated ORF. In this latter case, to do so, Mikado will check upstream and
        downstream to where the nearest in-frame start and stop codons are found.
        This function will error if expanding the ORF will lead to an in-frame stop codon upstream of the original
        start point.

        :param sequence: the cDNA sequence
        :type sequence: str
        :param upstream: how many bases we are expanding upstream
        :type upstream: int
        :param downstream: how many bases we are expanding downstream
        :type downstream: int
        :param expand_orf: boolean switch. If set to True, Mikado will check the ORF for in-frame stop codons.
        :type expand_orf: bool
        :param logger: a logger for the function
        :type logger: logging.Logger
        """

        if upstream < 0 or downstream < 0:
            raise ValueError("Upstream and downstream must be 0 or positive, not {upstream} and {downstream}".format(
                upstream=upstream, downstream=downstream))

        if len(sequence) != len(self) + upstream + downstream:
            raise ValueError(
                "When trying to expand the original sequence of length {lself} of {sid} by {upstream} upstream and {downstream} downstream nucleotides, \
the total length ({total}) is different from the length of the provided sequence ({lseq}).".format(
                    sid=self.id, lself=len(self), upstream=upstream, downstream=downstream, total=len(self) + upstream + downstream,
                    lseq=len(sequence)))

        if len(self) == len(sequence):
            logger.debug(
                "The length of the sequence for {sid} is identical to the length of the original object. No action needed.".format(
                    sid=self.id))
            return

        if self.transcriptomic is False:
            raise ValueError("This is not a transcriptomic BED12, I cannot expand it!\n{sself}".format(
                sself=repr(self)))
        if self.strand == "-":
            raise NotImplementedError(
                "{sid} is on the negative strand, I can only expand ORFs on the sense strand".format(sid=self.id))

        old_sequence = sequence[upstream:len(self) + upstream]
        if len(old_sequence) + upstream + downstream != len(sequence):
            raise ValueError(
                "When trying to expand the original sequence of length {lself} of {sid} by {upstream} upstream and {downstream} downstream nucleotides, \
the length of the *imputed* old sequence ({lold}) does not tally up with the new sequence ({lnew})".format(
                    lself=len(self), sid=self.id, upstream=upstream, downstream=downstream,
                    lold=len(old_sequence), lnew=len(sequence)))
        self.fasta_length = len(sequence)

        # I presume that the sequence is already in the right orientation
        old_start_pos = self.thick_start + self.phase - 1
        old_end_pos = self.thick_end - (self.thick_end - old_start_pos) % 3
        old_orf = old_sequence[old_start_pos:old_end_pos].upper()
        logger.debug("Old sequence of %s (%s bps): %s[...]%s", self.id, len(old_sequence),
                     old_sequence[:10], old_sequence[-10:])
        logger.debug("Old ORF of %s (%s bps, phase %s): %s[...]%s", self.id, len(old_orf), self.phase,
                     old_orf[:10], old_orf[-10:])
        # TODO: this function should not fail for non-coding transcripts
        assert len(old_orf) > 0, (old_start_pos, old_end_pos)
        assert len(old_orf) % 3 == 0, (old_start_pos, old_end_pos)

        old_pep = _translate_str(old_orf, self.table, gap="N")
        if "*" in old_pep and old_pep.find("*") < len(old_pep) - 1:
            logger.error("Stop codon found within the ORF of %s (pos %s of %s; phase %s). This is invalid!",
                         self.id, old_pep.find("*"), len(old_pep), self.phase)

        self.start_codon = old_orf[:3]
        self.stop_codon = old_orf[-3:]
        logger.debug("%s: start codon %s, old start %s (%s); stop codon %s, old stop %s (%s)",
                     self.name, self.start_codon, self.thick_start + self.phase,
                     (self.thick_start + self.phase + upstream),
                     self.stop_codon, self.thick_end, (self.thick_end + upstream))
        # Now expand
        self.end = len(sequence)
        self.thick_start += upstream
        self.thick_end += upstream
        start_codon = str(self.start_codon).upper()
        stop_codon = str(self.stop_codon).upper()
        self.has_start_codon = (start_codon in self.table.start_codons)
        self.has_stop_codon = (stop_codon in self.table.stop_codons)
        self.logger.debug("%s has start codon (%s): %s", self.chrom, start_codon, self.has_start_codon)
        self.logger.debug("%s has stop codon (%s): %s", self.chrom, stop_codon, self.has_stop_codon)
        if expand_orf is True and not (self.has_start_codon and self.has_stop_codon):
            if not self.has_start_codon:
                for pos in range(old_start_pos + upstream,
                                 0,
                                 -3):
                    codon = sequence[pos:pos + 3].upper()

                    self.thick_start = pos + 1
                    if codon in self.table.start_codons:
                        # self.thick_start = pos
                        self.start_codon = codon
                        self.__has_start = True
                        logger.debug("Position %d, codon %s. Start codon found.", pos, codon)
                        break
                if self.start_codon not in self.table.start_codons:
                    self.phase = (self.thick_start - 1) % 3
                    logger.debug("No start codon found for %s. Thick start %s, new phase: %s",
                                 self.id, self.thick_start, self.phase)
                    self.thick_start = 1
                else:
                    self.phase = 0
                    self.__has_start = True

            coding_seq = sequence[self.thick_start + self.phase - 1:self.end]
            if len(coding_seq) % 3 != 0:
                # Only get a multiple of three
                coding_seq = coding_seq[:-((len(coding_seq)) % 3)]
            prot_seq = _translate_str(coding_seq, table=self.table, gap="N")
            if "*" in prot_seq:
                self.thick_end = self.thick_start + self.phase - 1 + (1 + prot_seq.find("*")) * 3
                self.stop_codon = coding_seq[prot_seq.find("*") * 3:(1 + prot_seq.find("*")) * 3].upper()
                self.__has_stop = True
                logger.debug("New stop codon for %s: %s", self.name, self.thick_end)

            if self.stop_codon not in self.table.stop_codons:
                logger.debug("No valid stop codon found for %s", self.name)
                self.thick_end = self.end

        self.block_sizes = [self.thick_end - self.thick_start]
        self.block_starts = [self.thick_start]
        return

    @property
    def blocks(self):

        """This will return the coordinates of the blocks, with a 1-offset (as in GFF3)"""

        # First thing: calculate where each start point will be
        starts = self.block_starts + self.start - 1
        _bstarts = starts + 1
        _bends = starts + self.block_sizes

        return list(zip(_bstarts, _bends))

    def to_transcriptomic(self, sequence=None, fasta_index=None, start_adjustment=False,
                          lenient=False, alias=None, coding=True):

        """This method will return a transcriptomic version of the BED12. If the object is already transcriptomic,
        it will return itself."""

        if self.transcriptomic is True:
            return self
        self.coding = coding

        # Find the exon(s) in which the thick start and thick end are located and calculate their *transcriptomic*
        # coordinates
        tStart, tEnd = None, None
        seen = 0
        for block in self.blocks:
            if tStart and tEnd:
                # We have calculated them
                break
            if not tStart and block[0] <= self.thick_start <= block[1]:
                tStart = seen + self.thick_start - block[0] + 0  # adder[0]
            if not tEnd and block[0] <= self.thick_end <= block[1]:
                tEnd = seen + self.thick_end - block[0] + 1
            seen += block[1] - block[0] + 1

        # Check thick start and end are defined

        assert tStart is not None and tEnd is not None, f"The thick start, thick end of {self.id} are invalid " \
                                                        f"as they are outside of the defined exons.\nThick start: " \
                                                        f"{self.thick_start}\nThick end: {self.thick_end}\n" \
                                                        f"Exons: {self.blocks}"

        if self.strand == "+":
            bsizes = self.block_sizes[:]
        else:
            bsizes = np.flip(self.block_sizes)
            tStart, tEnd = self.block_sizes.sum() - tEnd, self.block_sizes.sum() - tStart

        bstarts = np.concatenate([np.zeros(1, dtype=np.int64), bsizes[:-1].cumsum()])
        if not (len(bstarts) == len(bsizes) == self.block_count):
            raise ValueError("""In {self.id} ({self.chrom}:{self.start}-{self.end}) there is a discrepancy between block \
starts (# {lbstarts}, {bstarts}), block sizes (# {lbsizes}, {bsizes}) and block counts (# {self.block_count}). \
This is invalid""".format(self=self, lbstarts=len(bstarts), lbsizes=len(bsizes), bstarts=bstarts, bsizes=bsizes))

        if self.coding:
            new_name = "ID={};coding={};phase={}".format(self.name.split(";")[0],
                                                         self.coding,
                                                         self.phase if self.phase is not None else 0)
        else:
            new_name = "ID={};coding={}".format(self.name.split(";")[0], self.coding)

        if alias is not None:
            new_name += ";alias={}".format(alias)

        if not self.coding:
            tStart, tEnd = 0, 1

        new = list((self.name.split(";")[0],
                    0,
                    self.block_sizes.sum(),
                    new_name,
                    self.score,
                    "+"))

        new.extend(list((
            tStart,
            tEnd,
            self.rgb,
            self.block_count,
            bsizes,
            bstarts
        )))

        new = BED12(new,
                    phase=self.phase,
                    sequence=sequence,
                    coding=self.coding,
                    fasta_index=fasta_index,
                    transcriptomic=True,
                    lenient=lenient,
                    start_adjustment=start_adjustment)
        assert isinstance(new, type(self)), f"The new object is of type {type(new)} instead of {type(self)}!"
        return new

    @property
    def logger(self):
        """
        Property. It returns the logger instance attached to the class.
        :rtype : logging.Logger | None
        """

        return self.__logger

    @logger.setter
    def logger(self, logger):
        """Set a logger for the instance.
        :param logger: a Logger instance
        :type logger: logging.Logger | None
        """
        if logger is None:
            if self.__logger is None:
                logger = create_null_logger()
                self.__logger = logger
            else:
                pass
        else:
            if not isinstance(logger, logging.Logger):
                raise TypeError(
                    "Objects of type {tself} accept only logging.Logger instances as loggers, not {tlog}!".format(
                        tself=type(self), tlog=type(logger)))
            self.__logger = logger
        self.__logger.propagate = False


class Bed12Parser(Parser):
    """Parser class for a Bed12Parser file.
    It accepts optionally a fasta index which is used to
    determine whether an ORF has start/stop codons."""

    __annot_type__ = "bed12"

    def __init__(self, handle,
                 fasta_index=None,
                 transcriptomic=False,
                 max_regression=0,
                 start_adjustment=True,
                 is_gff=False,
                 coding=False,
                 logger=create_null_logger(),
                 table=0):
        """
        Constructor method.
        :param handle: the input BED file.

        :param fasta_index: optional FAI file

        :param transcriptomic: flag. If set to True, it indicates
        that the BED file contains ORF information (like that derived
        from transdecoder) and is in transcriptomic rather than genomic coordinates.
        :type transcriptomic: bool

        :param max_regression: parameter to pass directly to BED12, indicating how much
        we should backtrack in the sequence to find a valid start codon
        """

        Parser.__init__(self, handle)
        self.transcriptomic = transcriptomic
        self.__max_regression = 0
        self._max_regression = max_regression
        self.coding = coding
        self.start_adjustment = start_adjustment
        self.logger = logger
        self.fasta_index = self.__set_fasta_index(fasta_index)
        self.__closed = False
        self.header = False
        self.__table = table
        self._is_bed12 = (not is_gff)
        self.__line_counter = 0

    @staticmethod
    def __set_fasta_index(fasta_index):
        if isinstance(fasta_index, dict):
            # check that this is a bona fide dictionary ...
            assert isinstance(
                fasta_index[random.choice(fasta_index.keys())],
                Bio.SeqRecord.SeqRecord)
        elif fasta_index is not None:
            if isinstance(fasta_index, (str, bytes)):
                if isinstance(fasta_index, bytes):
                    fasta_index = fasta_index.decode()
                assert os.path.exists(fasta_index)
                fasta_index = pysam.FastaFile(fasta_index)
            else:
                assert isinstance(fasta_index, pysam.FastaFile), type(fasta_index)
        return fasta_index

    def __iter__(self):
        return self

    def __next__(self, seq=None):

        try:
            if self._is_bed12 is True:
                return self.bed_next()
            else:
                return self.gff_next()
        except (ValueError, KeyError, TypeError, UnicodeError, AttributeError, AssertionError, InvalidParsingFormat) as exc:
            raise InvalidParsingFormat(f"This is not a valid BED12 file! Exception: {exc}")

    def __getstate__(self):
        state = super().__getstate__()
        # Now let's remove the fasta index
        if state["fasta_index"] is not None:
            if isinstance(state["fasta_index"], pysam.FastaFile):
                state["fasta_index"] = state["fasta_index"].filename
        return state

    def __setstate__(self, state):
        fasta_index = state.pop("fasta_index", None)
        super().__setstate__(state)
        self.logger = create_null_logger()
        self.__set_fasta_index(fasta_index)

    def bed_next(self):
        """

        :return:
        """

        bed12 = None
        while bed12 is None:
            line = next(self._handle)
            self.__line_counter += 1
            try:
                bed12 = BED12(line,
                              fasta_index=self.fasta_index,
                              transcriptomic=self.transcriptomic,
                              max_regression=self._max_regression,
                              coding=self.coding,
                              table=self.__table,
                              logger=self.logger,
                              start_adjustment=self.start_adjustment)
            except (ValueError, TypeError, CodonTable.TranslationError, KeyError, InvalidParsingFormat) as exc:
                error = "Invalid line for file {name}, line {counter}:\n{line}\nError: {exc}".format(
                    name=self.name, counter=self.__line_counter, line=line.rstrip(), exc=exc)
                raise InvalidParsingFormat(error)
        return bed12

    def gff_next(self):
        """

        :return:
        """

        bed12 = None
        while bed12 is None:
            line = next(self._handle)
            self.__line_counter += 1
            try:
                gff_line = GffLine(line)
            except (ValueError, TypeError, CodonTable.TranslationError, KeyError, InvalidParsingFormat) as exc:
                error = "Invalid line for file {name}, line {counter}:\n{line}\nError: {exc}".format(
                    name=self.name, counter=self.__line_counter, line=line.rstrip(), exc=exc)
                raise InvalidParsingFormat(error)

            if gff_line.feature != "CDS":
                continue
            # Compatibility with BED12
            try:
                bed12 = BED12(gff_line,
                              fasta_index=self.fasta_index,
                              transcriptomic=self.transcriptomic,
                              max_regression=self._max_regression,
                              table=self.__table,
                              start_adjustment=self.start_adjustment,
                              logger=self.logger)
            except (ValueError, TypeError, CodonTable.TranslationError, KeyError, InvalidParsingFormat) as exc:
                error = "Invalid line for file {name}, line {counter}:\n{line}\nError: {exc}".format(
                    name=self.name, counter=self.__line_counter, line=line.rstrip(), exc=exc)
                raise InvalidParsingFormat(error)
        # raise NotImplementedError("Still working on this!")
        return bed12

    def close(self):
        if self.__closed is False:
            self._handle.close()
            self.__closed = True

    @property
    def _max_regression(self):
        """
        This property is used to indicate how far downstream we should go in the
          FASTA sequence to find a valid start codon, in terms of percentage
          of the the cDNA sequence. So eg in a 300 nt cDNA a max_regression value
          of 0.3 would instruct the class to look only for the first 90 bps for
          a Met.
        """

        return self.__max_regression

    @_max_regression.setter
    def _max_regression(self, value):
        if not (isinstance(value, (int, float)) and 0 <= value <= 1):
            raise ValueError(
                "Invalid value specified for _max_regression (must be between 0 and 1): {}".format(value))
        self.__max_regression = value

    @property
    def coding(self):
        if not hasattr(self, "__coding"):
            self.__coding = False
        return self.__coding

    @coding.setter
    def coding(self, coding):
        if coding not in (False, True):
            raise ValueError(coding)
        self.__coding = coding


class Bed12ParseWrapper(mp.Process):

    def __init__(self,
                 identifier=None,
                 rec_queue=None,
                 return_queue=None,
                 log_queue=None,
                 level="DEBUG",
                 fasta_index=None,
                 transcriptomic=False,
                 max_regression=0,
                 is_gff=False,
                 coding=False,
                 start_adjustment=True,
                 table=0):

        """
        :param send_queue:
        :type send_queue: mp.Queue
        :param return_queue:
        :type send_queue: mp.Queue
        :param kwargs:
        """

        super().__init__()
        if identifier is None:
            raise ValueError
        self.__identifier = identifier
        self.rec_queue = rec_queue
        self.return_queue = return_queue
        self.logging_queue = log_queue
        self.transcriptomic = transcriptomic
        self.__max_regression = 0
        self._max_regression = max_regression
        self.coding = coding
        self.start_adjustment = start_adjustment

        if isinstance(fasta_index, dict):
            # check that this is a bona fide dictionary ...
            assert isinstance(
                fasta_index[random.choice(fasta_index.keys())],
                Bio.SeqRecord.SeqRecord)
        elif fasta_index is not None:
            if isinstance(fasta_index, (str, bytes)):
                if isinstance(fasta_index, bytes):
                    fasta_index = fasta_index.decode()
                assert os.path.exists(fasta_index)
                fasta_index = pyfaidx.Fasta(fasta_index)
            else:
                assert isinstance(fasta_index, pysam.FastaFile), type(fasta_index)
        self._level = level
        self.fasta_index = fasta_index
        self.__closed = False
        self.header = False
        self.__table = table
        self._is_bed12 = (not is_gff)

    def bed_next(self, line, sequence=None):
        """

        :return:
        """

        try:
            bed12 = BED12(line,
                          logger=self.logger,
                          sequence=sequence,
                          transcriptomic=self.transcriptomic,
                          max_regression=self._max_regression,
                          start_adjustment=self.start_adjustment,
                          coding=self.coding,
                          table=self.__table)
        except (ValueError, TypeError, CodonTable.TranslationError, KeyError, InvalidParsingFormat) as exc:
            raise InvalidParsingFormat("Invalid line: {}".format(line))
        return bed12

    def gff_next(self, line, sequence):
        """

        :return:
        """

        try:
            line = GffLine(line)
        except (ValueError, TypeError, CodonTable.TranslationError, KeyError, InvalidParsingFormat) as exc:
            error = "Invalid line:\n{}".format(line)
            raise InvalidParsingFormat(error)

        if line.feature != "CDS":
            return None
            # Compatibility with BED12
        bed12 = BED12(line,
                      logger=self.logger,
                      sequence=sequence,
                      transcriptomic=self.transcriptomic,
                      max_regression=self._max_regression,
                      start_adjustment=self.start_adjustment,
                      table=self.__table)
        # raise NotImplementedError("Still working on this!")
        return bed12

    def run(self, *args, **kwargs):
        self.handler = logging_handlers.QueueHandler(self.logging_queue)
        self.logger = logging.getLogger(self.name)
        self.logger.addHandler(self.handler)
        self.logger.setLevel(self._level)
        self.logger.propagate = False

        self.logger.info("Started %s", self.__identifier)
        if self.rec_queue is None:
            self.return_queue.put(b"FINISHED")
            raise ValueError
        while True:
            if self.rec_queue.empty():
                sleep(0.1)
                continue
            line = self.rec_queue.get()
            if line in ("EXIT", b"EXIT"):
                self.rec_queue.put(b"EXIT")
                self.return_queue.put(b"FINISHED")
                break
            try:
                num, line, seq = line
                if seq is not None:
                    seq = zlib.decompress(seq).decode()
                if not self._is_bed12:
                    row = self.gff_next(line, seq)
                else:
                    row = self.bed_next(line, seq)

                if not row or row.header is True:
                    continue
                if row.invalid is True:
                    self.logger.warning("Invalid entry, reason: %s\n%s",
                                        row.invalid_reason,
                                        row)
                    continue
                # self.cache[num] = 
                self.return_queue.put((num, msgpack.dumps(row.as_simple_dict())))
            except AttributeError:
                pass
            except (ValueError, TypeError, CodonTable.TranslationError, KeyError, InvalidParsingFormat) as exc:
                raise InvalidParsingFormat(line)



# coding: utf-8

"""
This module defines a child of the Transcript class, which is used
to verify that e.g. the assigned strand is correct.
"""
import functools

from .transcript import Transcript
from ..exceptions import IncorrectStrandError, InvalidTranscript, InvalidCDS
from collections import Counter
from itertools import zip_longest
from ..parsers.bed12 import BED12
import pyfaidx
from Bio import Seq


# pylint: disable=too-many-instance-attributes
class TranscriptChecker(Transcript):
    """This is a subclass of the generic transcript class. Its purpose is to compare
    the information of the transcript instance with the information contained in a
    genomic FASTA file, to verify some of the information.
    At the moment, the class implements only a check on the strandedness made by extracting
    the FASTA sequence of the splice sites and verifying that they are concordant
    with the annotated strand.
    Keyword arguments:
        - strand_specific: if set, monoexonic transcripts are not set to "unknown" strand.
        - lenient: boolean. If set to True, a transcript with mixed splices will
        not throw an exception but rather will just report the number
        of splices supporting each strand.
    """

    # The arguments are all needed here.
    # pylint: disable=too-many-arguments
    def __init__(self, gffline, seq,
                 strand_specific=False, lenient=False,
                 is_reference=False,
                 canonical_splices=(("GT", "AG"), ("GC", "AG"), ("AT", "AC")),
                 strip_faulty_cds=False,
                 logger=None):

        """
        Constructor method. It inherits from Transcript, with some modifications.

        :param gffline: annotation line to begin the construction.
        :type gffline: (Mikado.parsers.gfannotation.GFAnnotation|Transcript|Mikado.parsers.BED12|str)

        :param seq: a SeqIO indexed FASTA file

        :param strand_specific: flag. If set, transcripts will not have their strand changed.
        :type strand_specific: bool

        :param lenient: boolean flag. If set, incorrect transcripts will be flagged rather than discarded.
                        Please note that this flag will *not* be passed on to the BED12 constructor when checking the
                        ORFs; Mikado prepare requires correct CDSs, if one has to be present at all.
        :type lenient: bool
        """
        self.__strand_specific = False
        self.__is_reference = False
        self.mixed_attribute = ""

        if seq is None:
            raise ValueError("No sequence provided!")
        if isinstance(gffline, (Transcript, TranscriptChecker)):
            self.__dict__.update(gffline.__dict__)
        else:
            super().__init__(gffline)
        if self.strand is not None:
            self.original_strand = self.strand[:]
        else:
            self.original_strand = None
        self.__set_fasta_seq(seq)
        self.checked = False
        self.mixed_splices = False
        self.reversed = False
        self.canonical_splices = []
        self.__is_reference = False
        self.is_reference = is_reference
        self.strip_faulty_cds = strip_faulty_cds
        self.lenient = (self.is_reference or lenient)
        self.strand_specific = (self.is_reference or strand_specific)
        if not isinstance(canonical_splices, (tuple, list)):
            raise ValueError("Canonical splices should be provided as lists or tuples")

        for canonical_splice in canonical_splices:
            self.canonical_splices.append((str(canonical_splice[0]),
                                           str(canonical_splice[1])))

        self.canonical_junctions = []
        self.logger = logger

    # pylint: enable=too-many-arguments

    @property
    def fasta_seq(self):
        return self.__fasta_seq

    def __set_fasta_seq(self, fasta_seq):

        if isinstance(fasta_seq, Seq.Seq):
            self.__fasta_seq = pyfaidx.Sequence(name=self.id, seq=str(fasta_seq))
        elif isinstance(fasta_seq, str):
            self.__fasta_seq = pyfaidx.Sequence(name=self.id, seq=str(fasta_seq))
            assert len(self.__fasta_seq) == len(str(fasta_seq))
        elif isinstance(fasta_seq, pyfaidx.Sequence):
            self.__fasta_seq = fasta_seq
        else:
            raise ValueError("Unkown type: {}".format(type(fasta_seq)))

    @classmethod
    def rev_complement(cls, string):

        """
        Quick method to perform the reverse complement of a given string,
        using the class translation table.

        :param string: the sequence to be rev-complented
        :type string: str
        """

        return Seq.reverse_complement(string)

    @property
    def strand_specific(self):
        """
        Flag, set from the constructor. If True, transcript will not have their strand changed.
        :rtype: bool
        """
        return self.__strand_specific

    @strand_specific.setter
    def strand_specific(self, value):
        """
        Setter for strand_specific. Only boolean values are considered as valid.
        :param value: flag
        :type value: bool

        """
        if not isinstance(value, bool):
            raise TypeError("Invalid value for boolean property: {0}".format(value))
        self.__strand_specific = value

    def __str__(self, print_cds=True, to_gtf=False, with_introns=False):

        self.__prepare_for_printing()
        return super().__str__(print_cds=print_cds, to_gtf=to_gtf)

    def format(self, format_name, with_introns=False, with_cds=True,
               all_orfs=False,
               transcriptomic=False):

        self.__prepare_for_printing()
        return super().format(format_name, with_cds=with_cds, with_introns=with_introns,
                              all_orfs=all_orfs, transcriptomic=transcriptomic)

    def __prepare_for_printing(self):

        self.check_strand()
        self.check_orf()
        if self.mixed_splices is True:
            self.attributes["mixed_splices"] = self.mixed_attribute
            self.logger.warning("%s has mixed splices", self.id)
        if self.is_reference is True:
            self.attributes["is_reference"] = "True"
            self.logger.debug("%s is a reference transcript", self.id)

    def check_strand(self):
        """
        This method will check that the transcript instance has all the splice sites
        on one strand,or at most with non-canonical (therefore unknowable) splice junctions.
        If the transcript is monoexonic and strand_specific is set to False,
        the strand of the transcript will be set to None.

        The finalize method is called preliminarly before any operation.
        """

        self.finalize()

        if self.end - self.start + 1 != len(self.fasta_seq):
            error = """For {}, the expected length derived from last exon vs transcript start ({} vs {}) is different
            from the length of the FASTA sequence: {} vs {}""".format(self.id,
                                                                      self.exons[-1][1],
                                                                      self.start,
                                                                      self.exons[-1][1] - self.start + 1,
                                                                      len(self.fasta_seq))
            self.logger.error(error)
            raise InvalidTranscript(error)

        if self.checked is True:
            return

        if self.strand_specific is False and self.monoexonic is True and self.is_coding is False:
            self.strand = None

        elif self.monoexonic is False:
            canonical_counter = Counter()
            canonical_index = dict()

            for pos, intron in enumerate(self.introns):
                canonical_index[pos+1] = self._check_intron(intron)

            canonical_counter.update(canonical_index.values())

            if canonical_counter[None] == len(self.introns):
                self.logger.debug("Transcript %s only has non-canonical splices", self.id)
                if self.lenient is False:
                    raise IncorrectStrandError("No correct strand found for {0}".format(
                        self.id))

            elif canonical_counter["+"] > 0 and canonical_counter["-"] > 0:
                if self.lenient is False:
                    err_messg = """Transcript {0} has {1} positive and {2} negative splice junctions. \
    Aborting.""".format(self.id,
                        canonical_counter["+"],
                        canonical_counter["-"])
                    self.logger.warning(err_messg)
                    raise IncorrectStrandError(err_messg)

                else:
                    self.mixed_splices = True
                    self.logger.debug("Transcript %s has %d positive and %s negative splice junctions",
                                        self.id,
                                        canonical_counter["+"],
                                        canonical_counter["-"])

                    if canonical_counter["+"] >= canonical_counter["-"] or self.strand_specific is True:
                        self.mixed_attribute = "{0}concordant,{1}discordant".format(
                            canonical_counter["+"],
                            canonical_counter["-"])
                    else:
                        self.reverse_strand()
                        self.reversed = True
                        self.mixed_attribute = "{0}concordant,{1}discordant".format(
                            canonical_counter["-"],
                            canonical_counter["+"])

            elif canonical_counter["-"] > 0:
                if self.is_reference is False:
                    self.logger.debug("Transcript %s only has negative splice junctions, (%s), reversing.",
                                  self.id, canonical_counter)
                    self.reverse_strand()
                    self.reversed = True
                else:
                    self.logger.debug(
                        "Transcript %s only has negative splice junctions, but as it is a reference \
we will not reverse it")

            if canonical_counter["+"] >= canonical_counter["-"] or (
                            max(canonical_counter["-"], canonical_counter["+"]) > 0 and self.strand_specific is True):
                strand = "+"
            else:
                strand = "-"

            self.attributes["canonical_number"] = canonical_counter[strand]
            self.attributes["canonical_proportion"] = canonical_counter[strand] / len(self.introns)

            self.canonical_junctions = [_ for _ in canonical_index if
                                        canonical_index[_] == strand]
            self.attributes["canonical_junctions"] = ",".join([str(_) for _
                                                               in self.canonical_junctions])

        self.checked = True
        return

    def reverse_strand(self):
        if self.is_reference is True:
            raise InvalidTranscript("I cannot reverse the strand of a reference transcript.")
        elif self.is_coding is True and self.strip_faulty_cds is False:
            raise InvalidTranscript("I cannot reverse the strand of a coding transcript.")
        super().reverse_strand()

    def _check_intron(self, intron):

        """
        Private method that checks whether an intron has canonical splice sites
        or not.
        :param intron: the intron tuple (int,int) in 1-base offset
        [("AG","GT")]
        :return: strand of the intron (None | "+" | "-")
        """

        splice_donor = self.fasta_seq[intron[0] - self.start:intron[0]-self.start + 2]
        assert len(splice_donor) == 2
        splice_acceptor = self.fasta_seq[intron[1] - 1 - self.start:intron[1] - self.start + 1]

        if not isinstance(splice_donor, str):
            splice_donor = str(splice_donor.seq)
            splice_acceptor = str(splice_acceptor.seq)

        assert isinstance(splice_acceptor, str)

        splice_donor = splice_donor.upper()
        splice_acceptor = splice_acceptor.upper()

        if self.strand == "-":
            splice_donor, splice_acceptor = (self.rev_complement(splice_acceptor),
                                             self.rev_complement(splice_donor))
        if (splice_donor, splice_acceptor) in self.canonical_splices:
            strand = "+"
        else:
            splice_donor, splice_acceptor = (self.rev_complement(splice_acceptor),
                                             self.rev_complement(splice_donor))
            if (splice_donor, splice_acceptor) in self.canonical_splices:
                strand = "-"
            else:

                strand = None
        return strand

    def check_orf(self):

        """
        Method to check that the transcript ORF makes sense (no internal stop codons and the like).
        If it fails, the transcript will have its ORF removed.
        """

        self.has_start_codon = False
        self.has_stop_codon = False

        try:
            self.finalize()
        except InvalidCDS:
            if self.strip_faulty_cds is True:
                self.strip_cds(strand_specific=True)
                self.finalize()
            else:
                self.logger.warning("Transcript %s has an invalid CDS and must therefore be discarded.", self.id)
                raise InvalidTranscript("Transcript %s has an invalid CDS and must therefore be discarded.")

        if self.is_coding is False:
            return

        try:
            orfs = self.orfs
            if len(orfs) < 1:
                raise AssertionError("No ORFs remaining.")
        except AssertionError as exc:
            if self.strip_faulty_cds is False:
                raise InvalidTranscript(f"Invalid ORF(s) for {self.id}. Discarding it. Error: {exc}")
            else:
                self.logger.warning(f"Invalid ORF(s) for {self.id}. Stripping it of its CDS. Error: {exc}")
                self.strip_cds()
            return

        orfs = self.find_overlapping_cds(orfs)
        assert len(orfs) >= 1

        for orf_counter, orf in enumerate(orfs, start=1):
            assert isinstance(orf, BED12)
            if orf.invalid and orf_counter == len(orfs):
                if self.strip_faulty_cds is False:
                    raise InvalidTranscript("Invalid ORF(s) for %s. Discarding it. Reason: %s", self.id,
                                            orf.invalid_reason)
                else:
                    self.logger.warning("Invalid ORF for %s (reason: %s)", self.id, orf.invalid_reason)
                    self.strip_cds(self.strand_specific)
                return
            elif not orf.invalid:
                self.logger.debug("%s %s a valid start codon, %s a valid stop codon",
                                  self.id, "has" if orf.has_start_codon else "does not have",
                                  "has" if orf.has_stop_codon else "does not have")
                self.has_start_codon = orf.has_start_codon
                self.has_stop_codon = orf.has_stop_codon
                self.attributes["has_start_codon"] = orf.has_start_codon
                self.attributes["has_stop_codon"] = orf.has_stop_codon

        self.strip_cds()
        self.load_orfs(orfs)

    @property
    def cdna(self) -> str:

        """This property calculates the cDNA sequence of the transcript."""

        return self.calculate_cdna(str(self.fasta_seq.seq), tuple(self.exons), self.cdna_length, self.strand,
                                   self.start)

    @staticmethod
    @functools.lru_cache(maxsize=10, typed=True)
    def calculate_cdna(fasta_seq, exons, cdna_length, strand, tstart):
        sequence = ""
        for exon in exons:
            start = exon[0] - tstart
            end = exon[1] + 1 - tstart
            _ = fasta_seq[start:end]
            assert len(_) == end - start, (len(_), end - start)
            sequence += _

        assert len(sequence) == cdna_length, (len(sequence), cdna_length)
        if strand == "-":
            sequence = TranscriptChecker.rev_complement(sequence)
        return sequence

    @property
    def fasta(self):
        """
        This property calculates and returns the FASTA sequence (with header) associated with
        the transcript instance.
        The FASTA sequence itself will be formatted to be in lines with 60 characters
        (the standard).
        :return:
        """

        self.check_strand()
        fasta = [">{0}".format(self.id)]

        # assert len(sequence) == sum(exon[1] + 1 - exon[0] for exon in self.exons), (
        #     len(sequence), sum(exon[1] + 1 - exon[0] for exon in self.exons)
        # )

        sequence = self.grouper(self.cdna, 60)
        fasta.extend(sequence)
        fasta = "\n".join(fasta)
        return fasta

    @staticmethod
    def grouper(iterable, num, fillvalue=""):
        """Collect data into fixed-length chunks or blocks. From core documentation.

        :param iterable: the iterable to be considered for grouping.
        :param num: length of the chunks.
        :type num: int

        :param fillvalue: the default filler for missing positions while grouping.
        :type fillvalue: str
        """
        # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx"
        args = [iter(iterable)] * num
        return list("".join(x) for x in zip_longest(*args, fillvalue=fillvalue))



# coding: utf-8

"""
This class is used to store all the data calculated by the Assigner class, in order to produce
the RefMap/Stats files.
"""

import argparse
import collections
import logging
import operator
from logging import handlers as log_handlers
from .. import version
from ..transcripts import Transcript
from ..utilities.namespace import Namespace
from ..utilities import calc_f1, Interval
from .resultstorer import ResultStorer
from ..utilities import IntervalTree
import networkx as nx
import numpy as np


# noinspection PyPropertyAccess,PyPropertyAccess,PyPropertyAccess
# pylint: disable=too-many-instance-attributes
class Accountant:
    """This class stores the data necessary to calculate the final statistics
     - base and exon Sn/Sp/F1 etc."""

    def __init__(self, genes, args: argparse.Namespace, counter=None, fuzzymatch=0, load_ref=False):

        """Class constructor. It requires:
        :param genes: a dictionary
        :type genes: [dict|GeneDict]
        :param args: A namespace (like those provided by argparse)
        containing the parameters for the run.
        """

        self.args = args
        self.queue_handler = None
        self.logger = None
        self._counter = counter
        self.__setup_logger()
        self.logger.debug("Started with stat printing")

        self.introns = dict()
        self.exons = dict()
        self.starts = dict()
        self.ends = dict()
        self.intron_chains = dict()
        self.monoexonic_matches = (set(), set())
        self.ref_genes = dict()
        self.pred_genes = dict()
        if load_ref is True:
            self.logger.info("Starting loading the reference for the accountant %s", self._counter)
            self.__setup_reference_data(genes)
        self.self_analysis = False
        self.__fuzzymatch = fuzzymatch
        if hasattr(args, "self") and args.self is True:
            self.self_analysis = True

    def __setup_reference_data(self, genes):

        """
        Private method that prepares the reference data into the data structure
        that will be used to compare each prediction with a reference transcript.
        """

        for gene in genes:
            self.ref_genes[gene] = dict()
            for transcr in genes[gene]:
                # 0b000, indicating:
                # First bit: stringent match (100% F1)
                # Second bit: normal match (95% F1)
                # Third bit: lenient match (80% F1)
                self.ref_genes[gene][transcr.id] = 0b1000
                #                 self.ref_transcript_num+=1
                if transcr.chrom not in self.introns:
                    self.exons[transcr.chrom] = dict([("+", dict()), ("-", dict())])
                    self.starts[transcr.chrom] = dict([("+", dict()), ("-", dict())])
                    self.ends[transcr.chrom] = dict([("+", dict()), ("-", dict())])

                    self.introns[transcr.chrom] = dict([("+", dict()), ("-", dict())])
                    self.intron_chains[transcr.chrom] = dict([("+", dict()), ("-", dict())])
                if transcr.strand is None:
                    strand = "+"
                else:
                    strand = transcr.strand

                # 0b000001: in reference
                # 0b000010: in prediction
                # 0b000100: single
                # 0b001000: internal
                # 0b010000: border
                # 0b100000: single match lenient

                if transcr.exon_num == 1:
                    exon = tuple([transcr.exons[0][0], transcr.exons[0][1]])
                    self.exons[transcr.chrom][strand][exon] = 0b00000
                    self.exons[transcr.chrom][strand][exon] |= 0b1
                    self.exons[transcr.chrom][strand][exon] |= 0b100
                else:
                    self.__store_multiexonic_reference(transcr, strand)

        return

    def __store_multiexonic_reference(self, transcr, strand):

        """
        Private method to store a multiexonic reference transcript.
        :param transcr:
        :param strand:
        :return:
        """

        intron_chain = []
        for intron in sorted(transcr.introns):
            intron = tuple([intron[0], intron[1]])
            intron_chain.append(intron)
            if intron not in self.introns[transcr.chrom][strand]:
                self.introns[transcr.chrom][strand][intron] = [0, 0]
            self.introns[transcr.chrom][strand][intron][0] += 1
        intron_chain = tuple(intron_chain)
        if intron_chain not in self.intron_chains[transcr.chrom][strand]:
            self.intron_chains[transcr.chrom][strand][intron_chain] = [set(), set()]
        self.intron_chains[transcr.chrom][strand][intron_chain][0].add(transcr.id)

        for index, exon in enumerate(transcr.exons):
            exon = tuple([exon[0], exon[1]])
            if exon not in self.exons[transcr.chrom][strand]:
                self.exons[transcr.chrom][strand][exon] = 0b00000
            self.exons[transcr.chrom][strand][exon] |= 0b01
            if index == 0:
                self.exons[transcr.chrom][strand][exon] |= 0b010000
                self.starts[transcr.chrom][strand][exon[1]] = 0b01
            elif index == transcr.exon_num - 1:
                self.exons[transcr.chrom][strand][exon] |= 0b010000
                self.ends[transcr.chrom][strand][exon[0]] = 0b01
            else:
                self.exons[transcr.chrom][strand][exon] |= 0b01000

    def __setup_logger(self):

        """
        Private method to set up the logger using indications in the
        args namespace.
        """

        if hasattr(self.args, "log_queue"):
            # noinspection PyUnresolvedReferences
            self.queue_handler = log_handlers.QueueHandler(self.args.log_queue)
        else:
            self.queue_handler = logging.NullHandler
        if self._counter is None:
            self.logger = logging.getLogger("stat_logger")
        else:
            self.logger = logging.getLogger("stat_logger-{}".format(self._counter))
        self.logger.addHandler(self.queue_handler)
        # noinspection PyUnresolvedReferences
        if self.args.verbose:
            self.logger.setLevel(logging.DEBUG)
        else:
            self.logger.setLevel(logging.INFO)

        self.logger.propagate = False
        return

    # pylint: disable=too-many-locals
    def __retrieve_gene_stats(self, store_name="ref"):

        """
        This private method recovers from the pred/ref store the statistics regarding the
        number of genes and transcripts found and missed,
        both stringently and leniently.
        The keyword store_name, which must evaluate to either
        "ref"
        or
        "pred",
        indicates whether we are intersted in prediction of reference data.
        :param store_name: str
        :return: { "stringent": (transcript_stringent, gene_stringent),
                   "lenient": (transcript_lenient, gene_lenient),
                   "total": total_transcripts,
                   "private": (private_transcripts, private_genes)
                   }
        """

        found_transcripts_stringent = 0  # 100% match
        found_transcripts = 0  # 95% match
        found_transcripts_lenient = 0  # 80% match
        found_genes_stringent = 0  # 100% match
        found_genes = 0  # 95% match
        found_genes_lenient = 0  # 80% match
        private_genes = 0
        private_transcripts = 0
        total_transcripts = 0

        if store_name == "ref":
            store = self.ref_genes
        elif store_name == "pred":
            store = self.pred_genes
        else:
            raise ValueError("Invalid store selected: {0}".format(store_name))

        # 0b000, indicating:
        # First bit: stringent match (100% F1)
        # Second bit: normal match (95% F1)
        # Third bit: lenient match (80% F1)
        # Fourth bit: private

        for gene in store:
            gene_match = 0b000
            gene_not_found = 0b1
            for _, val in store[gene].items():
                total_transcripts += 1
                found_transcripts_lenient += (0b100 & val) >> 2
                found_transcripts += (0b10 & val) >> 1
                found_transcripts_stringent += 0b1 & val
                # Will evaluate to 1 if the transcript has at least one match
                private_transcripts += (0b1000 & val) >> 3
                gene_not_found &= (0b1000 & val) >> 3
                gene_match |= val

            found_genes_stringent += gene_match & 0b1
            # Shift back by 1
            found_genes += (gene_match & 0b10) >> 1
            # Shift back by 2
            found_genes_lenient += (gene_match & 0b100) >> 2
            private_genes += gene_not_found

        self.logger.debug("""Found %s transcripts:
        \tstringent\t%s
        \tlenient\t%s
        \ttotal\t%s""",
                          store_name,
                          found_transcripts_stringent,
                          found_transcripts_lenient,
                          total_transcripts)

        result = dict()
        result["stringent"] = (found_transcripts_stringent, found_genes_stringent)
        result["standard"] = (found_transcripts, found_genes)
        result["lenient"] = (found_transcripts_lenient, found_genes_lenient)
        result["total"] = total_transcripts
        result["private"] = (private_transcripts, private_genes)

        return result
    # pylint: enable=too-many-locals

    def __calculate_gene_stats(self):

        """
         Private method that calculates the number of found transcripts
         in the reference and prediction sets. Wrapper around a double
         call to __retrieve_gene_stats
         :return: double the result of __retrieve_gene_stats
         """

        ref_results = self.__retrieve_gene_stats(store_name="ref")

        pred_results = self.__retrieve_gene_stats(store_name="pred")

        result = {"ref": ref_results,
                  "pred": pred_results}

        return result

    def __calculate_chains_stats(self, chains):

        """
        This takes up a dictionary of the form (intron_chain: set(reference ids), set(prediction ids))

        :param chains:
        :return:
        """

        graph = nx.Graph()
        positions = IntervalTree()

        evaluator = dict()
        intron_chains_common_ref = set()
        intron_chains_common_pred = set()
        intron_chains_pred = 0
        intron_chains_ref = 0

        intron_chains_nonred = np.array([0, 0, 0])

        for chain in chains:
            ichain = IntervalTree.from_tuples(chain)
            evaluator[ichain] = chain
            graph.add_node(ichain)
            positions.insert(Interval(ichain.start, ichain.end, value=ichain))

        for ichain, chain in evaluator.items():
            found = positions.find(ichain.start, ichain.end)
            for ochain in found:
                if ichain.fuzzy_equal(ochain.value, self.__fuzzymatch):  # Define which chains are broadly similar
                    graph.add_edge(ichain, ochain.value)

        communities = nx.algorithms.connected_components(graph)
        for community in communities:  # Now we have to go back to the original sets ...
            ref_vals = set()
            pred_vals = set()
            for chain in community:
                ref_vals.update(chains[evaluator[chain]][0])
                pred_vals.update(chains[evaluator[chain]][1])
            if ref_vals and pred_vals:
                intron_chains_common_ref.update(ref_vals)
                intron_chains_common_pred.update(pred_vals)
                intron_chains_nonred[0] += 1

            if ref_vals:
                intron_chains_nonred[2] += 1
                intron_chains_ref += len(ref_vals)
            if pred_vals:
                intron_chains_nonred[1] += 1
                intron_chains_pred += len(pred_vals)

        intron_chains_common_ref = len(intron_chains_common_ref)
        intron_chains_common_pred = len(intron_chains_common_pred)
        intron_chains_red = np.array([intron_chains_common_pred, intron_chains_common_ref,
                                       intron_chains_pred, intron_chains_ref])

        return intron_chains_red, intron_chains_nonred

    def __calculate_intron_stats(self):

        """
        This private method calculates the raw numbers regarding intron and intron chain statistics.
        :return: result_dictionary
        :rtype: dict
        """

        intron_stats = [0, 0, 0, 0]  # Common - pred side, common - ref side, prediction, reference
        intron_nr_stats = [0, 0, 0]  # Common, prediction, reference

        splice_stats = [0, 0, 0]  # Common, prediction, reference

        # Common, prediction, reference
        intron_chains_red = np.array([0, 0, 0, 0])  # Common - pred side, common - ref side, prediction, reference
        intron_chains_nonred = np.array([0, 0, 0])

        for chrom in self.intron_chains:
            for strand in self.intron_chains[chrom]:
                splice_starts = [set(), set()]  # reference, prediction
                splice_ends = [set(), set()]
                intron_nrs = [collections.Counter(), collections.Counter()]
                for chain, intron_val in self.intron_chains[chrom][strand].items():  # Two sets, refs and preds
                    __ref_ic_val = len(intron_val[0])
                    __pred_ic_val = len(intron_val[1])
                    for intron in chain:
                        if __ref_ic_val:  # reference
                            splice_starts[0].add(intron[0])
                            splice_ends[0].add(intron[1])
                        if __pred_ic_val:  # prediction
                            splice_starts[1].add(intron[0])
                            splice_ends[1].add(intron[1])
                        intron_nrs[0][intron] += __ref_ic_val
                        intron_nrs[1][intron] += __pred_ic_val

                __red, __nonred = self.__calculate_chains_stats(self.intron_chains[chrom][strand])
                intron_chains_red += __red
                intron_chains_nonred += __nonred

                splice_stats[0] += len(set.intersection(*splice_starts)) + len(set.intersection(*splice_ends))
                splice_stats[1] += len(splice_starts[1]) + len(splice_ends[1])  # prediction
                splice_stats[2] += len(splice_starts[0]) + len(splice_ends[0])  # reference
                for key in set.union(set(intron_nrs[0].keys()), set(intron_nrs[1].keys())):
                    __ref_val, __pred_val = intron_nrs[0][key], intron_nrs[1][key]
                    if min(__ref_val, __pred_val) > 0:
                        intron_stats[1] += __ref_val
                        intron_stats[0] += __pred_val
                        intron_nr_stats[0] += 1

                    intron_stats[2] += __pred_val
                    intron_stats[3] += __ref_val
                    intron_nr_stats[1] += (__pred_val > 0)
                    intron_nr_stats[2] += (__ref_val > 0)

        self.logger.debug("Intron stats: {}".format(", ".join([str(_) for _ in intron_stats])))
        # intron_nr_stats = [len(set.intersection(*intron_nrs)), len(intron_nrs[1]), len(intron_nrs[0])]
        self.logger.debug("Intron NR stats: {}".format(", ".join([str(_) for _ in intron_nr_stats])))


        self.logger.debug("""Intron chains:\
        \treference %d\t%d
        \tprediction\t%d
        \t%d
        \tcommon\t%d %d %d""",
                          intron_chains_red[3],
                          intron_chains_nonred[2],
                          intron_chains_red[2],
                          intron_chains_nonred[1],
                          intron_chains_nonred[0],
                          intron_chains_red[1],
                          intron_chains_red[0])

        result_dictionary = dict()
        result_dictionary["introns"] = dict()
        result_dictionary["introns"]["non_redundant"] = intron_nr_stats
        result_dictionary["introns"]["redundant"] = intron_stats
        result_dictionary["splices"] = splice_stats
        result_dictionary["intron_chains"] = dict()
        result_dictionary["intron_chains"]["non_redundant"] = intron_chains_nonred

        result_dictionary["intron_chains"]["redundant"] = intron_chains_red
        return result_dictionary

    def __extract_terminal_stats(self, result_dictionary):
        """
        Private method to
        :param result_dictionary: the results dictionary
        :type result_dictionary: dict
        :return: updated result dictionary
        :rtype: dict
        """

        # 0b000001: in reference
        # 0b000010: in prediction
        # 0b000100: single
        # 0b001000: internal
        # 0b010000: border
        # 0b100000: single match lenient

        exon_common_lenient = 0
        exon_ref_lenient = 0
        exon_pred_lenient = 0

        for chrom in self.starts:
            for strand in self.starts[chrom]:
                for start in self.starts[chrom][strand]:
                    exon_common_lenient += (0b1 & self.starts[chrom][strand][start]) & \
                                           ((0b10 & self.starts[chrom][strand][start]) >> 1)
                    exon_ref_lenient += 0b01 & self.starts[chrom][strand][start]
                    exon_pred_lenient += (0b10 & self.starts[chrom][strand][start]) >> 1
        starts_common = exon_common_lenient
        starts_ref = exon_ref_lenient
        starts_pred = exon_pred_lenient

        self.logger.debug("Starts %s", [starts_common, starts_ref, starts_pred])

        for chrom in self.ends:
            for strand in self.ends[chrom]:
                for end in self.ends[chrom][strand]:
                    exon_common_lenient += (0b01 & self.ends[chrom][strand][end]) & \
                                           ((0b10 & self.ends[chrom][strand][end]) >> 1)
                    exon_ref_lenient += 0b01 & self.ends[chrom][strand][end]
                    exon_pred_lenient += (0b10 & self.ends[chrom][strand][end]) >> 1

        ends_common = exon_common_lenient - starts_common
        ends_ref = exon_ref_lenient - starts_ref
        ends_pred = exon_pred_lenient - starts_pred
        self.logger.debug("Ends %s", [ends_common, ends_ref, ends_pred])
        result_dictionary["ends"] = [ends_common, ends_pred, ends_ref]
        result_dictionary["starts"] = [starts_common, starts_pred, starts_ref]
        result_dictionary["exons"] = dict()
        result_dictionary["exons"]["lenient"] = [exon_common_lenient,
                                                 exon_pred_lenient,
                                                 exon_ref_lenient]

        return result_dictionary

    def __extract_internal_stats(self, result_dictionary):
        """
        Private method that extracts the internal stats from the stored
        data and updates the result_dictionary dictionary.
        :param result_dictionary: the dictionary that stores the results
        :return:
        """

        # Re-extract the previous stats
        # Common, prediction, reference
        exon_lenient = result_dictionary["exons"]["lenient"]

        # Common, prediction, reference
        bases = [0, 0, 0]

        # Common, prediction, reference
        exon_stringent = [0, 0, 0]

        for chrom in self.exons:
            for strand in self.exons[chrom]:
                # 0b000001: in reference
                # 0b000010: in prediction
                # 0b000100: single
                # 0b001000: internal
                # 0b010000: border
                # 0b100000: single match lenient
                curr_pred_bases = set()
                curr_ref_bases = set()
                curr_exon = (-1, -1)

                for exon in sorted(self.exons[chrom][strand], key=operator.itemgetter(0)):
                    # Out of previous overlap
                    if exon[0] > curr_exon[1]:
                        bases[0] += len(set.intersection(curr_pred_bases, curr_ref_bases))
                        bases[1] += len(curr_pred_bases)
                        bases[2] += len(curr_ref_bases)
                        curr_exon = exon
                        curr_pred_bases = set()
                        curr_ref_bases = set()
                    else:
                        curr_exon = (curr_exon[0], exon[1])

                    # Exon is in reference
                    if (0b01 & self.exons[chrom][strand][exon]) == 0b1:
                        curr_ref_bases.update(set(range(exon[0], exon[1])))
                        exon_stringent[2] += 1
                        # Internal (first condition)
                        # OR
                        # Single exon
                        if (0b001000 & self.exons[chrom][strand][exon]) == 0b001000:
                            exon_lenient[2] += 1
                        elif 0b100000 & self.exons[chrom][strand][exon]:
                            exon_lenient[2] += 1

                    # Exon is in prediction
                    if (0b10 & self.exons[chrom][strand][exon]) == 0b10:
                        curr_pred_bases.update(set(range(exon[0], exon[1])))
                        exon_stringent[1] += 1
                        # Either internal, or a single exon which
                        # has not a lenient single match with the reference annotation
                        if (0b001000 & self.exons[chrom][strand][exon]) == 0b001000:
                            exon_lenient[1] += 1
                        elif 0b100000 & self.exons[chrom][strand][exon]:
                            exon_lenient[1] += 1

                    # In both
                    if (0b11 & self.exons[chrom][strand][exon]) == 0b11:
                        exon_stringent[0] += 1
                        if (0b001000 & self.exons[chrom][strand][exon]) == 0b001000:
                            exon_lenient[0] += 1
                        elif 0b100000 & self.exons[chrom][strand][exon]:
                            exon_lenient[0] += 1

                bases[0] += len(set.intersection(curr_pred_bases, curr_ref_bases))
                bases[1] += len(curr_pred_bases)
                bases[2] += len(curr_ref_bases)

        assert bases[0] <= min(bases[1], bases[2]), bases
        result_dictionary["bases"] = bases
        result_dictionary["exons"]["stringent"] = exon_stringent
        result_dictionary["exons"]["lenient"] = exon_lenient
        return result_dictionary

    def __calculate_exon_stats(self):

        """
        This private method calculates the raw numbers regarding base and exon statistics.
        :return:
        """

        result = dict()
        result = self.__extract_terminal_stats(result)
        result = self.__extract_internal_stats(result)
        return result

    def serialize(self):

        simplified = Namespace()
        simplified.attributes = ("pred_genes", "ref_genes", "intron_chains", "monoexonic_matches",
                                 "introns", "exons", "starts", "ends")

        for attr in simplified.attributes:
            setattr(simplified, attr, getattr(self, attr))

        return simplified

    def merge_into(self, accountant):

        for parent in accountant.pred_genes:
            if parent not in self.pred_genes:
                self.pred_genes[parent] = accountant.pred_genes[parent].copy()
            else:
                for tid in accountant.pred_genes[parent]:
                    if tid in self.pred_genes[parent]:
                        self.pred_genes[parent][tid] |= accountant.pred_genes[parent][tid]
                    else:
                        self.pred_genes[parent][tid] = accountant.pred_genes[parent][tid]

        def clearBit(int_type, offset):
            mask = ~(1 << offset)
            return (int_type & mask)

        def getBit(int_type, offset):
            mask = 1 << offset
            if (int_type & mask):
                return 1
            else:
                return 0

        def setBit(int_type, offset):
            mask = 1 << offset
            return (int_type | mask)

        for ref_gene in accountant.ref_genes:
            for refid in accountant.ref_genes[ref_gene]:
                val = self.ref_genes[ref_gene][refid]
                oval = accountant.ref_genes[ref_gene][refid]
                for pos in range(3):
                    bit = getBit(val, pos) | getBit(oval, pos)
                    if bit:
                        val = setBit(val, pos)
                bit = getBit(val, 3) & getBit(oval, 3)
                if not bit:
                    val = clearBit(val, 3)
                self.ref_genes[ref_gene][refid] = val

        for chrom in accountant.intron_chains:
            if chrom not in self.intron_chains:
                self.intron_chains[chrom] = dict()
            for strand in accountant.intron_chains[chrom]:
                if strand not in self.intron_chains[chrom]:
                    self.intron_chains[chrom][strand] = dict()
                for ic_key in accountant.intron_chains[chrom][strand]:
                    if ic_key not in self.intron_chains[chrom][strand]:
                        self.intron_chains[chrom][strand][ic_key] = [set(), set()]
                    self.intron_chains[chrom][strand][ic_key][0].update(
                        accountant.intron_chains[chrom][strand][ic_key][0]
                    )
                    self.intron_chains[chrom][strand][ic_key][1].update(
                        accountant.intron_chains[chrom][strand][ic_key][1]
                    )

        self.monoexonic_matches[0].update(accountant.monoexonic_matches[0])
        self.monoexonic_matches[1].update(accountant.monoexonic_matches[1])

        for feature in ("exons", "starts", "ends"):
            store = getattr(accountant, feature)
            my_store = getattr(self, feature)
            for chrom in store:
                if chrom not in my_store:
                    my_store[chrom] = dict()
                for strand in store[chrom]:
                    if strand not in my_store[chrom]:
                        my_store[chrom][strand] = dict()
                    for feat in store[chrom][strand]:
                        if feat not in my_store[chrom][strand]:
                            my_store[chrom][strand][feat] = 0b0
                        my_store[chrom][strand][feat] |= store[chrom][strand][feat]

    def __store_multiexonic_result(self, transcr, strand):

        """
        Private procedure to store the results regarding a multiexonic
        transcript compared to the reference.
        :param transcr: a transcript instance
        :param strand: the strand to be used for storing
        :return:
        """

        # assert transcr.monoexonic is False
        ic_key = tuple([tuple([intron[0], intron[1]]) for intron in sorted(transcr.introns)])
        # We register the variables here because multiple access slows the program down!
        chrom = transcr.chrom[:]
        exon_num = transcr.exon_num
        if ic_key not in self.intron_chains[chrom][strand]:
            self.intron_chains[chrom][strand][ic_key] = [set(), set()]
        self.intron_chains[chrom][strand][ic_key][1].add(transcr.id)

        for index, exon in enumerate(transcr.exons):
            exon = tuple([exon[0], exon[1]])
            if exon not in self.exons[chrom][strand]:
                self.exons[chrom][strand][exon] = 0b0
            self.exons[chrom][strand][exon] |= 0b10  # set it as "in prediction"
            if index == 0:
                self.exons[chrom][strand][exon] |= 0b010000
                if exon[1] not in self.starts[transcr.chrom][strand]:
                    self.starts[chrom][strand][exon[1]] = 0b0
                self.starts[chrom][strand][exon[1]] |= 0b10
            elif index == exon_num - 1:
                self.exons[chrom][strand][exon] |= 0b010000
                if exon[0] not in self.ends[chrom][strand]:
                    self.ends[chrom][strand][exon[0]] = 0b00
                self.ends[chrom][strand][exon[0]] |= 0b10
            else:
                self.exons[chrom][strand][exon] |= 0b01000

    def __store_monoexonic_result(self, transcr, strand, result: ResultStorer, other_exon=None):
        """
        Private procedure to store the comparison of a monoexonic
        transcript with the reference.
        :param transcr: the monoexonic transcript
        :param strand: the strand to be used for storing
        :param other_exon: the reference exon this transcript is assigned to
        :return:
        """
        assert transcr.monoexonic is True
        exon = tuple([transcr.exons[0][0], transcr.exons[0][1]])
        if exon not in self.exons[transcr.chrom][strand]:
            self.exons[transcr.chrom][strand][exon] = 0b0
        self.exons[transcr.chrom][strand][exon] |= 0b10
        self.exons[transcr.chrom][strand][exon] |= 0b100
        if result.ccode == ("_",):
            self.monoexonic_matches[0].add(result.ref_id)
            self.monoexonic_matches[1].add(transcr.id)

        if other_exon is not None:
            # other_exon = tuple([other_exon[0], other_exon[1]])
            assert isinstance(other_exon, tuple)
            if other_exon not in self.exons[transcr.chrom][strand]:
                self.exons[transcr.chrom][strand][other_exon] = 0b1  # Necessary as we are trying to avoid preloading

            assert other_exon in self.exons[transcr.chrom][strand]

            if not 0b100 & self.exons[transcr.chrom][strand][other_exon]:
                # Check the other exon is marked as single
                self.exons[transcr.chrom][strand][other_exon] |= 0b100

            # self.exons[transcr.chrom][strand][exon] |= 0b100000
            self.exons[transcr.chrom][strand][other_exon] |= 0b100000

    def store(self, transcr: Transcript, result: ResultStorer, other_exon):

        """Add exons introns intron chains etc. to the storage.
        :param transcr: a "Transcript" instance
        :param result: Instance of "result_storer"
        :param other_exon: either None or an integer duplex
        :type other_exon: None | (int,int)

        A transcript is considered a perfect match if it has:
            junction_f1==100 and nucleotide_f1==100;
        a lenient match if it has junction_f1==100 and nucleotide_f1>95,
        i.e. min(nucleotide_precision, nucleotide_recall)>90.4
        """

        for parent in transcr.parent:
            if parent not in self.pred_genes:
                self.pred_genes[parent] = dict()
            if transcr.id not in self.pred_genes[parent]:
                self.pred_genes[parent][transcr.id] = 0b1000

        if transcr.strand is None:
            strand = "+"
        else:
            strand = transcr.strand

        if result.ccode != ("u",):
            found_one = False

            for refid, refgene, nf1 in zip(result.ref_id, result.ref_gene, result.n_f1):
                if refgene not in self.ref_genes:
                    self.ref_genes[refgene] = dict()
                if refid not in self.ref_genes[refgene]:
                    self.ref_genes[refgene][refid] = 0b1000

                if nf1 > 0:
                    self.ref_genes[refgene][refid] &= ~(1 << 3)  # Clear the fourth bit
                    found_one = True
            if found_one:
                for parent in transcr.parent:
                    self.pred_genes[parent][transcr.id] &= ~(1 << 3)  # Clear the fourth bit

            if len(result.j_f1) == 1:
                zipper = zip(result.ref_id, result.ref_gene, result.j_f1, result.n_f1)
                for refid, refgene, junc_f1, nucl_f1 in zipper:
                    if junc_f1 == 100 and nucl_f1 >= 80:
                        for parent in transcr.parent:
                            self.pred_genes[parent][transcr.id] |= 0b100
                            if nucl_f1 >= 95:
                                self.pred_genes[parent][transcr.id] |= 0b10
                            if nucl_f1 == 100:
                                self.pred_genes[parent][transcr.id] |= 0b01
                        # Unset the "private" mark

                        self.ref_genes[refgene][refid] |= 0b100
                        if nucl_f1 >= 95:
                            self.ref_genes[refgene][refid] |= 0b10
                        if nucl_f1 == 100:
                            self.ref_genes[refgene][refid] |= 0b01
        else:
            for parent in transcr.parent:
                self.pred_genes[parent][transcr.id] &= 0b1000

        if transcr.chrom not in self.exons:
            self.exons[transcr.chrom] = dict([("+", dict()), ("-", dict())])
            self.starts[transcr.chrom] = dict([("+", dict()), ("-", dict())])
            self.ends[transcr.chrom] = dict([("+", dict()), ("-", dict())])
            self.introns[transcr.chrom] = dict([("+", dict()), ("-", dict())])
            self.intron_chains[transcr.chrom] = dict([("+", dict()), ("-", dict())])

        if transcr.exon_num > 1:
            self.__store_multiexonic_result(transcr, strand)
        else:
            self.__store_monoexonic_result(transcr, strand, result, other_exon=other_exon)

    @staticmethod
    def __calculate_statistics(common, pred, ref):
        """
        Function to calculate precision, recall and F1 given
        the numbers in common, of the prediction, and of the reference
        :param common:
        :param pred:
        :param ref:
        :return: precision, recall, f1stat
        """

        if ref > 0:
            recall = common / ref
        else:
            recall = 0
        if pred > 0:
            precision = common / pred
        else:
            precision = 0
        f1stat = calc_f1(recall, precision)

        return precision, recall, f1stat

    @staticmethod
    def __calculate_redundant_statistics(common_pred, common_ref, pred, ref):
        """
        Function to calculate precision, recall and F1 for non-redundant statistics given
        the numbers in common for the prediction, for the reference, and then the numbers of the reference and
        of the prediction.
        :param common:
        :param pred:
        :param ref:
        :return: precision, recall, f1stat
        """

        if ref > 0:
            recall = common_ref / ref
        else:
            recall = 0
        if pred > 0:
            precision = common_pred / pred
        else:
            precision = 0
        f1stat = calc_f1(recall, precision)

        return precision, recall, f1stat

    @staticmethod
    def __redundant_stats(common_pred, common_ref, pred, ref):

        """
        Function to calculate precision, recall and F1 given the numbers of:
        - predictions which are matches
        - reference which are matches
        - total predictions
        - total references

        :param common_pred: number of predictions which are a match
        :param common_ref: number of references which are a match
        :param pred: total number of predictions
        :param ref: total number of references
        :return:
        """

        if ref > 0:
            recall = common_ref / ref
        else:
            recall = 0

        if pred > 0:
            precision = common_pred / pred
        else:
            precision = 0

        f1stat = calc_f1(recall, precision)

        return precision, recall, f1stat

    @staticmethod
    def __format_rowname(stru):
        """
        Private method to format the name of the rows in the stats table
        :param stru:
        :return:
        """

        total_length = 35
        return " " * (total_length-len(stru)-1) + "{}:".format(stru.rstrip(":"))

    @classmethod
    def __format_comparison_line(cls, name, private, total):
        """
        Method to create the string that will be printed out to screen for the stats file.
        :return:
        """

        if total > 0:
            perc = 100 * private / total
        else:
            perc = 0

        string = "{0} {1}/{2}  ({3:.2f}%)".format(cls.__format_rowname(name),
                                                  private,
                                                  total,
                                                  perc)
        return string

    @classmethod
    def __format_comparison_segments(cls, name, private, common):
        """
        Method to create the string that will be printed out to screen for the stats file.
        :return:
        """

        if common > 0:
            perc = 100 * (private - common) / private
        else:
            perc = 0

        string = "{0} {1}/{2}  ({3:.2f}%)".format(cls.__format_rowname(name),
                                                  private-common,
                                                  private,
                                                  perc)
        return string

    def print_stats(self):

        """
        This method calculates and prints the final stats file.
        It is called when the prediction file
        parsing has been terminated correctly.
        """

        # num_ref_transcripts = sum(len(self.ref_genes[gid]) for gid in self.ref_genes)
        # num_pred_transcripts = sum(len(self.pred_genes[gid]) for gid in self.pred_genes)

        # Dictionary with the necessary data
        if self.self_analysis is True:
            self.logger.info("No general statistics file will be printed for a self-analysis run.")
            self.logger.removeHandler(self.queue_handler)
            return

        bases_exon_result = self.__calculate_exon_stats()
        intron_results = self.__calculate_intron_stats()
        gene_transcript_results = self.__calculate_gene_stats()

        # self.logger.debug("Missed exons lenient: %s", bases_exon_result["missed"])
        self.logger.debug("Exon stringent %s", bases_exon_result["exons"]["stringent"])
        self.logger.debug("Exon lenient %s", bases_exon_result["exons"]["lenient"])

        self.logger.debug(
            """Bases:
            \tcommon\t%d
            \tprediction\t%d
            \treference\t%d""",
            *bases_exon_result["bases"])

        bases_prec, bases_recall, bases_f1 = self.__calculate_statistics(
            *bases_exon_result["bases"])

        (exon_stringent_precision, exon_stringent_recall, exon_stringent_f1) = (
            self.__calculate_statistics(*bases_exon_result["exons"]["stringent"]))

        (exon_lenient_precision, exon_lenient_recall, exon_lenient_f1) = (
            self.__calculate_statistics(*bases_exon_result["exons"]["lenient"]))

        (splice_precision, splice_recall, splice_f1) = self.__calculate_statistics(
            *intron_results["splices"]
        )

        intron_precision, intron_recall, intron_f1 = self.__calculate_redundant_statistics(
            *intron_results["introns"]["redundant"]
        )

        intron_nr_precision, intron_nr_recall, intron_nr_f1 = self.__calculate_statistics(
            *intron_results["introns"]["non_redundant"]
        )

        # TODO: revise
        intron_chain_stats = self.__calculate_redundant_statistics(
            *intron_results["intron_chains"]["redundant"])

        intron_chains_precision, intron_chains_recall, intron_chains_f1 = intron_chain_stats

        intron_chain_nr_stats = self.__calculate_statistics(
            *intron_results["intron_chains"]["non_redundant"])

        intron_chains_nr_precision, intron_chains_nr_recall, intron_chains_nr_f1 = intron_chain_nr_stats

        # Transcript level
        # noinspection PyTypeChecker
        (tr_precision_stringent, tr_recall_stringent, tr_f1_stringent) = (
            self.__redundant_stats(
                gene_transcript_results["pred"]["stringent"][0],
                gene_transcript_results["ref"]["stringent"][0],
                gene_transcript_results["pred"]["total"],
                gene_transcript_results["ref"]["total"]))

        (tr_precision_standard, tr_recall_standard, tr_f1_standard) = (
            self.__redundant_stats(
                gene_transcript_results["pred"]["standard"][0],
                gene_transcript_results["ref"]["standard"][0],
                gene_transcript_results["pred"]["total"],
                gene_transcript_results["ref"]["total"]))

        # noinspection PyTypeChecker
        (tr_precision_lenient, tr_recall_lenient, tr_f1_lenient) = (
            self.__redundant_stats(
                gene_transcript_results["pred"]["lenient"][0],
                gene_transcript_results["ref"]["lenient"][0],
                gene_transcript_results["pred"]["total"],
                gene_transcript_results["ref"]["total"]))

        # Gene level
        ref_genes = len(self.ref_genes)
        pred_genes = len(self.pred_genes)
        # noinspection PyTypeChecker
        (gene_precision_stringent, gene_recall_stringent, gene_f1_stringent) = (
            self.__redundant_stats(
                gene_transcript_results["pred"]["stringent"][1],
                gene_transcript_results["ref"]["stringent"][1],
                pred_genes,
                ref_genes))

        # noinspection PyTypeChecker
        (gene_precision_standard, gene_recall_standard, gene_f1_standard) = (
            self.__redundant_stats(
                gene_transcript_results["pred"]["standard"][1],
                gene_transcript_results["ref"]["standard"][1],
                pred_genes,
                ref_genes))

        # noinspection PyTypeChecker
        (gene_precision_lenient, gene_recall_lenient, gene_f1_lenient) = (
            self.__redundant_stats(
                gene_transcript_results["pred"]["lenient"][1],
                gene_transcript_results["ref"]["lenient"][1],
                pred_genes,
                ref_genes))

        # noinspection PyUnresolvedReferences

        with open("{0}.stats".format(self.args.out), 'wt') as out:

            # noinspection PyUnresolvedReferences
            print("Mikado version {1}; Command line:\n{0:>10}".format(self.args.commandline,
                                                                      version.__version__), file=out)
            print(gene_transcript_results["ref"]["total"], "reference RNAs in",
                  len(self.ref_genes), "genes", file=out)
            print(gene_transcript_results["pred"]["total"], "predicted RNAs in ",
                  len(self.pred_genes), "genes", file=out)

            print("-" * 33, "|   Sn |   Pr |   F1 |", file=out)
            print("{0} {1:.2f}  {2:.2f}  {3:.2f}".format(
                self.__format_rowname("Base level"),
                bases_recall * 100,
                bases_prec * 100,
                bases_f1 * 100), file=out)
            print("{0} {1:.2f}  {2:.2f}  {3:.2f}".format(
                self.__format_rowname("Exon level (stringent)"),
                exon_stringent_recall * 100,
                exon_stringent_precision * 100,
                exon_stringent_f1 * 100), file=out)
            print("{0} {1:.2f}  {2:.2f}  {3:.2f}".format(
                self.__format_rowname("Exon level (lenient)"),
                exon_lenient_recall * 100,
                exon_lenient_precision * 100,
                exon_lenient_f1 * 100), file=out)
            print("{0} {1:.2f}  {2:.2f}  {3:.2f}".format(
                self.__format_rowname("Splice site level"),
                splice_recall * 100,
                splice_precision * 100,
                splice_f1 * 100), file=out)
            print("{0} {1:.2f}  {2:.2f}  {3:.2f}".format(
                self.__format_rowname("Intron level"),
                intron_recall * 100,
                intron_precision * 100, intron_f1 * 100), file=out)
            print("{0} {1:.2f}  {2:.2f}  {3:.2f}".format(
                self.__format_rowname("Intron level (NR)"),
                intron_nr_recall * 100,
                intron_nr_precision * 100, intron_nr_f1 * 100), file=out)
            print("{0} {1:.2f}  {2:.2f}  {3:.2f}".format(
                self.__format_rowname("Intron chain level"),
                intron_chains_recall * 100,
                intron_chains_precision * 100,
                intron_chains_f1 * 100), file=out)
            print("{0} {1:.2f}  {2:.2f}  {3:.2f}".format(
                self.__format_rowname("Intron chain level (NR)"),
                intron_chains_nr_recall * 100,
                intron_chains_nr_precision * 100,
                intron_chains_nr_f1 * 100), file=out)
            print("{0} {1:.2f}  {2:.2f}  {3:.2f}".format(
                self.__format_rowname("Transcript level (stringent)"),
                tr_recall_stringent * 100,
                tr_precision_stringent * 100,
                tr_f1_stringent * 100), file=out)
            print("{0} {1:.2f}  {2:.2f}  {3:.2f}".format(
                self.__format_rowname("Transcript level (>=95% base F1)"),
                tr_recall_standard * 100,
                tr_precision_standard * 100,
                tr_f1_standard * 100), file=out)
            print("{0} {1:.2f}  {2:.2f}  {3:.2f}".format(
                self.__format_rowname("Transcript level (>=80% base F1)"),
                tr_recall_lenient * 100,
                tr_precision_lenient * 100,
                tr_f1_lenient * 100), file=out)
            print("{0} {1:.2f}  {2:.2f}  {3:.2f}".format(
                self.__format_rowname("Gene level (100% base F1)"),
                gene_recall_stringent * 100,
                gene_precision_stringent * 100,
                gene_f1_stringent * 100), file=out)
            print("{0} {1:.2f}  {2:.2f}  {3:.2f}".format(
                self.__format_rowname("Gene level (>=95% base F1)"),
                gene_recall_standard * 100,
                gene_precision_standard * 100,
                gene_f1_standard * 100), file=out)
            print("{0} {1:.2f}  {2:.2f}  {3:.2f}".format(
                self.__format_rowname("Gene level (>=80% base F1)"),
                gene_recall_lenient * 100,
                gene_precision_lenient * 100,
                gene_f1_lenient * 100), file=out)
            print(file=out)
            print("#   Matching: in prediction; matched: in reference.", file=out)
            print("", file=out)
            print("{0} {1}".format(
                self.__format_rowname("Matching intron chains"),
                intron_results["intron_chains"]["redundant"][0]), file=out)
            print("{0} {1}".format(
                self.__format_rowname("Matched intron chains"),
                intron_results["intron_chains"]["redundant"][1]), file=out)
            print("{0} {1}".format(
                self.__format_rowname("Matching monoexonic transcripts"),
                len(self.monoexonic_matches[1])), file=out)
            print("{0} {1}".format(
                self.__format_rowname("Matched monoexonic transcripts"),
                len(self.monoexonic_matches[0])), file=out)
            print("{0} {1}".format(
                self.__format_rowname("Total matching transcripts"),
                intron_results["intron_chains"]["redundant"][0] + len(self.monoexonic_matches[1])),
                file=out)
            print("{0} {1}".format(
                self.__format_rowname("Total matched transcripts"),
                intron_results["intron_chains"]["redundant"][1] + len(self.monoexonic_matches[0])),
                file=out)

            print("", file=out)

            print(self.__format_comparison_segments("Missed exons (stringent)",
                                                    bases_exon_result["exons"]["stringent"][2],
                                                    bases_exon_result["exons"]["stringent"][0]),
                  file=out)
            print(self.__format_comparison_segments("Novel exons (stringent)",
                                                    bases_exon_result["exons"]["stringent"][1],
                                                    bases_exon_result["exons"]["stringent"][0]),
                  file=out)

            print(self.__format_comparison_segments("Missed exons (lenient)",
                                                    bases_exon_result["exons"]["lenient"][2],
                                                    bases_exon_result["exons"]["lenient"][0]),
                  file=out)
            print(self.__format_comparison_segments("Novel exons (lenient)",
                                                    bases_exon_result["exons"]["lenient"][1],
                                                    bases_exon_result["exons"]["lenient"][0]),
                  file=out)

            print(self.__format_comparison_segments("Missed introns",
                                                    intron_results["introns"]["non_redundant"][2],
                                                    intron_results["introns"]["non_redundant"][0]),
                  file=out)
            print(self.__format_comparison_segments("Novel introns",
                                                    intron_results["introns"]["non_redundant"][1],
                                                    intron_results["introns"]["non_redundant"][0]),
                  file=out)
            print("", file=out)

            # noinspection PyTypeChecker
            print(self.__format_comparison_line("Missed transcripts (0% nF1)",
                                                gene_transcript_results["ref"]["private"][0],
                                                gene_transcript_results["ref"]["total"]),
                  file=out)

            # noinspection PyTypeChecker
            print(self.__format_comparison_line("Novel transcripts (0% nF1)",
                                                gene_transcript_results["pred"]["private"][0],
                                                gene_transcript_results["pred"]["total"]),
                  file=out)

            # noinspection PyTypeChecker
            print(self.__format_comparison_line("Missed genes (0% nF1)",
                                                gene_transcript_results["ref"]["private"][1],
                                                len(self.ref_genes)), file=out)

            # noinspection PyTypeChecker
            print(self.__format_comparison_line("Novel genes (0% nF1)",
                                                gene_transcript_results["pred"]["private"][1],
                                                len(self.pred_genes)), file=out)

        self.logger.removeHandler(self.queue_handler)
        # self.queue_handler.close()
        return



#!/usr/bin/env python3

from Mikado.loci import Transcript, Gene
import argparse
import sys

from Mikado.parsers import parser_factory
from Mikado.parsers.GFF import GFF3

__doc__ = """This script truncates transcript with UTR exons separated by long introns."""

from Mikado.utilities import Interval


def remove_introns_from_transcr(transcript, args):
    transcript.finalize()
    if transcript.monoexonic is True:
        return transcript
    to_remove = set()

    cstart, cend = sorted([transcript.combined_cds_start, transcript.combined_cds_end])
    candidate_exons = [exon for exon in transcript.exons if
                       exon[1] < transcript.combined_cds_start or
                       exon[0] > transcript.combined_cds_end]
    
    # print(candidate_exons)
    if len(candidate_exons) == 0:
        return transcript
    
    for intron in transcript.introns - transcript.combined_cds_introns:
        if intron[1] - intron[0] + 1 >= args.max_intron:
            # print(transcript.id, "intron", intron, intron[1] - intron[0] + 1, file=sys.stderr)
            if intron[1] < cstart:
                to_remove.update([exon for exon in candidate_exons if exon[1] < intron[0]])
            elif intron[0] > cend:
                to_remove.update([exon for exon in candidate_exons if exon[0] > intron[1]])
            else:
                raise ValueError

    if len(to_remove) > 0:
        # cds = [_ for _ in transcript.segments if _[0] == "CDS"]
        # transcript.strip_cds()
        # transcript.combined_utr = []
        # transcript.segments = []
        transcript.finalized = False
        # transcript.internal_orfs[0].extend(cds)
        for exon in to_remove:
            transcript.remove_exon(Interval(exon[0], exon[1]))
        # print(cds)
        # for exon in cds:
        #     transcript.add_exon(exon, "CDS")
        # transcript.add_exons([(_[1][0], _[1][1]) for _ in cds], features="CDS")
        # transcript.internal_orfs[0].extend(cds)
        transcript.start = min(_[0] for _ in transcript.exons)
        transcript.end = max(_[1] for _ in transcript.exons)
        # print(transcript.internal_orfs)
        transcript.finalize()
        assert transcript.combined_cds_length > 0
    return transcript

def remove_introns(current, args):
    assert len(current.transcripts) > 0
    for tid in current.transcripts:
        transcript = current.transcripts[tid]
        if transcript.monoexonic:
            continue
        current.transcripts[tid] = remove_introns_from_transcr(transcript, args)
        # assert isinstance(current.transcripts[tid], Transcript)

    assert len(current.transcripts) > 0
    current.start = min(_.start for _ in current)
    current.end = max(_.end for _ in current)
    current.finalize()
    return current
    

def main():

    parser = argparse.ArgumentParser(__doc__)
    parser.add_argument("-mi", "--max-intron", default=10000, dest="max_intron",
                        type=int, help="Maximum intron length for UTR introns.")
    parser.add_argument("gff", type=parser_factory)
    parser.add_argument("out", default=sys.stdout, type=argparse.FileType("wt"), nargs="?")
    args = parser.parse_args()

    if args.max_intron < 0:
        raise ValueError("Max intron length <0 specified! {0}".format(args.max_intron))

    ref_gff = isinstance(args.gff, GFF3)
    if ref_gff:
        form = "gff3"
    else:
        form = "gtf"

    current = None
    current_transcript = None

    last_header = []
    for record in args.gff:
        if record.header is True:
            # print(record, file=sys.stderr)
            if current is not None:
                current = remove_introns(current, args)
                print(current.format(form), file=args.out)
                print(*last_header, sep="\n", file=args.out, end='')
                current = None
                current_transcript = None
            print(*last_header, sep="\n", end="")
            last_header = [record]
            continue
        if record.feature not in ("gene", "mRNA", "CDS", "exon"):
            continue
        
        if record.is_gene is True and ref_gff:
            print(record, file=sys.stderr)
            last_header = []
            if current is not None:
                # current = remove_introns(current, args)
                print(current.format(form), file=args.out)
                print(*last_header, sep="\n", file=args.out, end='')
                current = None
                current_transcript = None
        if record.is_transcript:
            if ref_gff is False:
                if current_transcript is not None:
                    # current_transcript = remove_introns_from_transcr(current_transcript,
                    #                                                 args)
                    assert current_transcript.combined_cds_length > 0
                    print(current_transcript, file=args.out)
                    print(*last_header, sep="\n", file=args.out)
                    last_header = []
            elif ref_gff is True:
                if current_transcript is not None:
                    if current is None:
                        current = Gene(current_transcript)
                        current.add(current_transcript)
                    else:
                        assert current_transcript.parent[0] != current.id
                        current.add(current_transcript)
                    # if current.id == current_transcript.parent[0]:
                        
                    # else:
                    #     current = remove_introns(current, args)
                    #     print(current.format(form), file=args.out)
                    #     print("###", file=args.out)
                    #     current = None
                
                # elif current_transcript is not None:
                #     current = Gene(current_transcript)

            current_transcript = Transcript(record)
        elif record.is_exon:
            if record.feature not in ("CDS", "exon"):
                continue
            current_transcript.add_exon(record)
        else:
            continue
        continue
    
    if ref_gff and current is not None:
        print(*last_header, sep="\n", file=args.out)
        last_header = []
        current = remove_introns(current, args)
        print(current.format(form), file=args.out)
    elif not ref_gff and current_transcript is not None:
        current_transcript = remove_introns_from_transcr(current_transcript,
                                                         args)
        print(current_transcript.format(form), file=args.out)
        print(*last_header, sep="\n", file=args.out, end='')
        
main()



from ..loci import Superlocus
import sys


def _create_locus_lines(stranded_locus: Superlocus,
                        gene_counter: int):
    configuration = stranded_locus.configuration

    if not stranded_locus.loci:
        return None, gene_counter

    for locus in stranded_locus.loci:
        gene_counter += 1
        new_id = "{0}.{1}G{2}".format(
            configuration.pick.output_format.id_prefix,
            stranded_locus.chrom, gene_counter)
        stranded_locus.loci[locus].id = new_id

    if len(stranded_locus.loci) > 0:
        assert stranded_locus.start != sys.maxsize
        assert stranded_locus.end != -sys.maxsize

    if stranded_locus.start != sys.maxsize:
        assert not stranded_locus.id.endswith("{0}--{0}".format(sys.maxsize))

    locus_lines = stranded_locus.__str__(
        print_cds=not configuration.pick.run_options.exclude_cds,
        level="loci")
    locus_metrics_rows = [x for x in stranded_locus.print_loci_metrics()]
    locus_scores_rows = [x for x in stranded_locus.print_loci_scores()]
    batch = [locus_lines, locus_metrics_rows, locus_scores_rows]
    return batch, gene_counter



# coding: utf-8

"""
Pretty basic class that defines a reference gene with its transcripts.
Minimal checks.
"""

import re
import copy
import logging
import operator
from sys import intern
from typing import Union, Iterable

from .transcript import Transcript
from .transcriptcomputer import TranscriptComputer
from ..exceptions import InvalidTranscript, InvalidCDS
from ..parsers.GFF import GffLine
from ..parsers.GTF import GtfLine
from ..utilities.log_utils import create_null_logger


class Gene:

    """
    :param transcr: a transcript used to initialize the container.
    :param gid:Id of the gene.
    :param logger: an optional Logger from the logging module.
    """

    __name__ = "gene"

    def __init__(self, transcr: Union[None, Transcript], gid=None, logger=create_null_logger(),
                 only_coding=False, use_computer=False):

        """Class describing a generic Gene object.
        :param transcr: either None or the transcript to initialise the Gene with.
        :type transcr: (None|Transcript)
        :param gid: the gene ID
        :type transcr: (None|str)
        :type logger: (None|logging.Logger)
        :param only_coding: boolean flag. If True, only accept coding transcripts
        :param use_computer: boolean flag. If True, use the TranscriptComputer class instead of the vanilla Transcript
        class. This is necessary for calculating statistics (see `mikado util stats`)
        """


        self.transcripts = dict()
        self.__logger = None
        self.logger = logger
        self.__introns = None
        self.exception_message = ''
        self.chrom, self.source, self.start, self.end, self.strand = [None] * 5
        self.only_coding = only_coding
        self.coding_transcripts = set()
        self.id = None
        self.attributes = dict()
        self.feature = "gene"
        self.__from_gene = False
        self.__use_computer = use_computer

        if transcr is not None:
            if isinstance(transcr, Transcript):
                if isinstance(transcr, TranscriptComputer):
                    self.__use_computer = True
                else:
                    self.__use_computer = False

                self.transcripts[transcr.id] = transcr
                if transcr.parent:
                    self.id = transcr.parent[0]
                else:
                    self.logger.debug("No gene ID found for %s, creating a mock one.", transcr.id)
                    transcr.parent = f"{transcr.id}.gene"
                    self.id = transcr.parent[0]
                self.transcripts[transcr.id] = transcr
            elif isinstance(transcr, GffLine):
                if transcr.is_gene is True:
                    self.__from_gene = True
                    self.id = transcr.id
                    self.attributes = transcr.attributes.copy()
                    self.feature = transcr.feature
                elif transcr.is_exon is True:
                    self.__from_gene = False
                    if transcr.parent:
                        self.id = transcr.parent[0]
                    else:
                        self.id = transcr.id
                    self.attributes = transcr.attributes.copy()
                    self.feature = "gene"
                    if self._use_computer is True:
                        ntranscr = TranscriptComputer(transcr,
                                                      trust_orf=True,
                                                      logger=self.logger,
                                                      accept_undefined_multi=True,
                                                      source=transcr.source,
                                                      is_reference=True)
                    else:
                        ntranscr = Transcript(transcr,
                                              trust_orf=True,
                                              logger=self.logger,
                                              accept_undefined_multi=True,
                                              source=transcr.source,
                                              is_reference=True)
                    ntranscr.add_exon(transcr)
                    self.add(ntranscr)
                    self.logger.debug("New transcript for %s: %s", self.id, ntranscr.id)
            elif isinstance(transcr, GtfLine):
                self.id = transcr.gene

            self.chrom, self.source, self.start, self.end, self.strand = (transcr.chrom,
                                                                          transcr.source,
                                                                          transcr.start,
                                                                          transcr.end,
                                                                          transcr.strand)
        if gid is not None:
            self.id = gid
        # Internalize in memory for less memory usage
        [intern(str(_)) for _ in [self.chrom, self.source, self.id]
         if _ is not None]

    def __contains__(self, item):

        if isinstance(item, (str, bytes)):
            return item in self.transcripts
        elif isinstance(item, Transcript):
            return item in self.transcripts.values()

    def keys(self):
        """Keys of the transcript dictionary"""
        return self.transcripts.keys()

    @property
    def logger(self):

        """
        Logger instance for the class.
        :rtype : logging.Logger
        """
        return self.__logger

    @logger.setter
    def logger(self, logger):
        """Set a logger for the instance.
        :param logger
        :type logger: logging.Logger | None
        """
        if isinstance(logger, logging.Logger):
            self.__logger = logger
        elif logger is None:
            name = "gene_{0}".format(self.id if self.id else "generic")
            self.__logger = create_null_logger(name)
        else:
            raise TypeError("Invalid object for logger: {0}, (type {1})".format(
                logger, type(logger)))

        for tid in self.transcripts:
            self.transcripts[tid].logger = logger

    @logger.deleter
    def logger(self):
        """
        Destroyer for the logger. It sets the internal __logger attribute to None.
        """
        self.__logger = None

    def add(self, transcr: Transcript):
        """
        This method adds a transcript to the storage.
        :param transcr: the transcript to be added.
        """

        self.transcripts[transcr.id] = transcr
        if self.chrom is None:
            self.chrom = transcr.chrom
        elif self.chrom != transcr.chrom:
            raise AssertionError("Discrepant chromosome for gene {0} and transcript {1}".format(
                    self.id, transcr.id
                ))

        if transcr.strand != self.strand:
            if self.strand is None:
                self.strand = transcr.strand
            elif transcr.strand is None:
                transcr.strand = self.strand
            else:
                raise AssertionError("Discrepant strands for gene {0} ({2}) and transcript {1} ({3})".format(
                    self.id, transcr.id, self.strand, transcr.strand
                ))

        transcr.logger = self.logger

    _prot_pattern = re.compile("-Protein$")

    def add_exon(self, row):
        """
        Add an exon to the transcripts of the class. The exon should have clearly marked its origin.
        :param row:
        :type row: (GtfLine | GffLine)
        :return:
        """

        found_tids = set()
        for parent in (_ for _ in row.parent if _ in self.transcripts):
            found_tids.add(parent)
            self.transcripts[parent].add_exon(row)

        for parent in (_ for _ in row.parent if _ not in self.transcripts):
            found = False
            if self._prot_pattern.search(parent) and self._prot_pattern.sub(r"", parent):
                continue
            
            for tid in self.transcripts:
                if parent in self.transcripts[tid].derived_children:
                    found = True
                    if tid not in found_tids:
                        self.transcripts[tid].add_exon(row)
                    break
            if not found:
                if self._use_computer is True:
                    self.transcripts[parent] = TranscriptComputer(row,
                                                                  trust_orf=True,
                                                                  logger=self.logger,
                                                                  accept_undefined_multi=True,
                                                                  source=row.source,
                                                                  is_reference=True,
                                                                  )
                else:
                    self.transcripts[parent] = Transcript(row,
                                                          trust_orf=True,
                                                          logger=self.logger,
                                                          accept_undefined_multi=True,
                                                          source=row.source,
                                                          is_reference=True,
                                                          )
                self.transcripts[parent].parent = self.id

        if row.id in self.transcripts:
            found_tids.add(row.id)
            self.transcripts[row.id].add_exon(row)

        self.logger.debug("Found transcripts: %s", found_tids)

    def __getitem__(self, tid: str) -> Transcript:
        return self.transcripts[tid]

    def finalize(self, exclude_utr=False):
        """
        This method will finalize the container by checking the consistency of all the
        transcripts and eventually removing incorrect ones.

        :param exclude_utr: boolean flag
        :return:
        """

        to_remove = set()
        for tid in self.transcripts:
            try:
                self.transcripts[tid].finalize()
                if self.only_coding is True and self.transcripts[tid].selected_cds_length == 0:
                    to_remove.add(tid)
                if self.transcripts[tid].selected_cds_length > 0:
                    self.coding_transcripts.add(tid)
                if exclude_utr is True:
                    self.transcripts[tid].remove_utrs()
            except InvalidCDS:
                self.transcripts[tid].strip_cds()
            except InvalidTranscript as err:
                self.exception_message += "{0}\n".format(err)
                to_remove.add(tid)
            except Exception as err:
                self.exception_message += "Error in gene {} for transcript {}".format(self.id, tid)
                self.exception_message += "{0}\n".format(err)
                self.logger.exception(self.exception_message)
                raise

        for k in to_remove:
            del self.transcripts[k]

        if len(self.transcripts) > 0:
            if self.source is None:
                self.source = set([_.source for _ in self.transcripts.values()]).pop()
            __new_start = min(_.start for _ in self)

            if __new_start != self.start:
                if self.__from_gene is True:
                    self.logger.warning("Resetting the start for %s from %s to %d",
                                        self.id, self.start, __new_start)
                else:
                    self.logger.debug("Resetting the start for %s from %s to %d",
                                      self.id, self.start, __new_start)

                self.start = __new_start

            __new_end = max(_.end for _ in self)
            if __new_end != self.end:
                if self.__from_gene is True:
                    self.logger.warning("Resetting the end for %s from %s to %d",
                                        self.id, self.end, __new_end)
                else:
                    self.logger.debug("Resetting the end for %s from %s to %d",
                                      self.id, self.end, __new_end)
                self.end = __new_end
        if self.exception_message:
            self.logger.exception(self.exception_message)
            self.exception_message = ""

    def as_dict(self, remove_attributes=True):

        """
        Method to transform the gene object into a JSON-friendly representation.
        :return:
        """

        state = dict()
        for key in ["chrom", "source", "start", "end", "strand", "id"]:
            state[key] = getattr(self, key)

        state["use_computer"] = self._use_computer
        state["transcripts"] = dict.fromkeys(self.transcripts.keys())

        for tid in state["transcripts"]:
            try:
                state["transcripts"][tid] = self.transcripts[tid].as_dict(
                    remove_attributes=remove_attributes)
            except TypeError as exc:
                self.logger.error(tid, repr(self.transcripts[tid]))
                self.logger.exception(exc)
                raise TypeError(exc)

        return state

    def load_dict(self, state: dict, exclude_utr=False, protein_coding=False, trust_orf=False):

        """
        Method to reconstitute a Gene object from a dictionary/JSON-like dump
        :param state: Gene dump
        :type state: dict
        :param exclude_utr: whether to remove UTRs from incoming transcripts.
        :param protein_coding: Whether to only consider coding transcripts for this gene
        :param trust_orf: parameter to be passed to Transcript.load_dict
        :return:
        """

        for key in ["chrom", "source", "start", "end", "strand", "id"]:
            setattr(self, key, state[key])

        self.__use_computer = state["use_computer"]
        for tid, tvalues in state["transcripts"].items():
            if self._use_computer is True:
                transcript = TranscriptComputer(logger=self.logger)
            else:
                transcript = Transcript(logger=self.logger)
            transcript.load_dict(tvalues, trust_orf=trust_orf)
            transcript.finalize()
            if protein_coding is True and transcript.is_coding is False:
                self.logger.debug("{0} is non coding ({1})".format(
                    transcript.id,
                    transcript.combined_cds,
                    ))
                continue
            if exclude_utr is True:
                has_utrs = (transcript.utr_length > 0)
                transcript.remove_utrs()
                if has_utrs is True and (transcript.utr_length > 0):
                    raise AssertionError("Failed to remove the UTRs!")
            self.transcripts[tid] = transcript

        self.chrom = intern(self.chrom)
        if self.source:
            self.source = intern(self.source)
        self.id = intern(self.id)

        return

    def remove(self, tid: str):
        """

        :param tid: name of the transcript to remove.

        This method will remove a transcript from the container, and recalculate the
         necessary instance attributes.

        """

        del self.transcripts[tid]
        if len(self.transcripts) == 0:
            self.end = None
            self.start = None
            self.chrom = None
        else:
            self.start = min(self.transcripts[tid].start for tid in self.transcripts)
            self.end = max(self.transcripts[tid].end for tid in self.transcripts)

    def __repr__(self):
        return " ".join(self.transcripts.keys())

    def __str__(self):
        return self.format("gff3")

    def __iter__(self) -> Iterable[Transcript]:
        """Iterate over the transcripts attached to the gene."""
        return iter(self.transcripts.values())

    def __len__(self) -> int:
        return len(self.transcripts)

    def __getstate__(self):

        logger = self.logger
        del self.logger
        state = self.__dict__.copy()
        self.logger = logger
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.logger = None

    def __lt__(self, other):
        if self.chrom != other.chrom:
            return self.chrom < other.chrom
        else:
            if self.start != other.start:
                return self.start < other.start
            elif self.end != other.end:
                return self.end < other.end
            else:
                if self.strand is not None and other.strand is not None:
                    return self.strand < other.strand
                elif self.strand is None and other.strand is not None:
                    return False
                elif self.strand is not None and other.strand is None:
                    return True
                else:
                    return False

    def __eq__(self, other):
        if self.chrom == other.chrom and self.start == other.start and \
                self.end == other.end and self.strand == other.strand:
            return True
        return False

    def format(self, format_name, transcriptomic=False):

        """Method for creating the necessary lines to be printed.
        Accepted formats are gff, gff3 (equivalent), bed12, bed (equivalent), gtf.

        :param format_name: the name of the format to create
        :type format_name: str
        :param transcriptomic: whether to create lines in genomic (False) or transcriptomic (True) coordinates.
        :type transcriptomic: bool

        """

        if format_name not in ("gff", "gtf", "gff3", "bed12", "bed"):
            raise ValueError(
                "Invalid format: {0}. Accepted formats: bed/bed12 (equivalent), gff/gff3 (equivalent), gtf".format(
                    format_name))

        self.finalize()  # Necessary to sort the exons
        lines = []
        if format_name in ("gff", "gff3") and transcriptomic is False:
            line = GffLine(None)
            for attr in ["chrom",
                         "source",
                         "start",
                         "end",
                         "strand"]:
                setattr(line, attr, getattr(self, attr))
            line.feature = self.feature
            line.attributes = self.attributes.copy()
            line.id = self.id
            assert line.id is not None
            lines.append(str(line))

        for tid, transcript in sorted(self.transcripts.items(), key=operator.itemgetter(1)):
            lines.append(transcript.format(format_name, transcriptomic=transcriptomic))

        if format_name in ("gff", "gff3"):
            lines.append("###")
        return "\n".join(lines)

    def copy(self):

        """Method to return a deep copy of the gene."""

        logger = self.logger
        del self.logger
        state = copy.deepcopy(self)
        self.logger = logger
        return state

    @property
    def monoexonic(self):
        """
        Boolean property. False if at least one transcript is multiexonic,
        True otherwise.
        :return: bool
        :rtype: bool
        """

        return all(transcript.monoexonic is True for transcript in self.transcripts.values())

    @property
    def introns(self):
        """
        It returns the set of all introns in the container.
        :rtype : set
        """

        return set.union(*[self.transcripts[tid].introns for tid in self.transcripts])

    @property
    def exons(self):
        """
        It returns the set of all exons in the container.
        :rtype : set
        """
        return set.union(*[set(self.transcripts[tid].exons) for tid in self.transcripts])

    @property
    def has_monoexonic(self):
        """
        True if any of the transcripts is monoexonic.
        :rtype : bool
        """
        return any(len(self.transcripts[tid].introns) == 0 for tid in self.transcripts.keys())

    # @property
    # def monoexonic(self):
    #     return all(len(self.transcripts[tid].introns) == 0 for tid in self.transcripts.keys())

    @property
    def num_transcripts(self):
        """
        Number of transcripts.
        :rtype : int
        """
        return len(self.transcripts)

    @property
    def num_coding_transcripts(self):
        """
        Number of coding transcripts
        :return:
        """

        return len([_ for _ in self.transcripts if self.transcripts[_].is_coding is True])

    @property
    def is_coding(self):
        """
        Property. It evaluates to True if at least one transcript is coding, False otherwise.
        """
        return any(self.transcripts[tid].selected_cds_length > 0 for tid in self.transcripts.keys())

    @property
    def _use_computer(self):
        return self.__use_computer